"""Can we parse UGC strings"""

import pytest
from pyiem.exceptions import UGCParseException
from pyiem.util import utc
from pyiem.nws import ugc

STR1 = "DCZ001-170200-"
STR2 = "DCZ001-MDZ004>007-009>011-013-014-016>018-VAZ036>042-050>057-170200-"


def test_missed_ugc():
    """Invalid encoded county string, check that NMC006 was not included"""
    text = (
        "NMC001-005>007-011-019-027-028-031-033-039-041-043-045-"
        "047-049-053-055-057-061-040300-"
    )
    valid = utc(2018, 6, 4, 3, 0)
    (ugcs, expire) = ugc.parse(text, valid)
    assert expire == valid
    assert ugcs[3] != ugc.UGC("NM", "C", "006")


def test_louisana():
    """Test that some specific logic works."""
    ugcs = [ugc.UGC("LA", "C", i) for i in range(100)]
    res = ugc.ugcs_to_text(ugcs)
    ans = "100 parishes in [LA]"
    assert res == ans


def test_totextstr():
    """ See if we can generate a proper string from a UGCS """
    ugcs = [
        ugc.UGC("DC", "Z", "001"),
        ugc.UGC("IA", "C", "001"),
        ugc.UGC("IA", "C", "002"),
    ]
    assert (
        ugc.ugcs_to_text(ugcs)
        == "((DCZ001)) [DC] and ((IAC001)), ((IAC002)) [IA]"
    )


def test_parse_exception():
    """Test that we raise a proper exception when given bad data."""
    valid = utc(2008, 12, 17, 3, 0)
    text = "IA078-170300-"
    with pytest.raises(UGCParseException):
        ugc.parse(text, valid)


def test_str1():
    """ check ugc.parse of STR1 parsing """
    valid = utc(2008, 12, 17, 3, 0)
    (ugcs, expire) = ugc.parse(STR1, valid)

    expire_answer = valid.replace(hour=2)
    ugcs_answer = [ugc.UGC("DC", "Z", "001")]

    assert ugcs == ugcs_answer
    assert expire == expire_answer


def test_str2():
    """ check ugc.parse of STR2 parsing """
    valid = utc(2008, 12, 17, 3, 0)
    (ugcs, expire) = ugc.parse(STR2, valid)

    expire_answer = valid.replace(hour=2)
    ugcs_answer = [
        ugc.UGC("DC", "Z", "001"),
        ugc.UGC("MD", "Z", "004"),
        ugc.UGC("MD", "Z", "005"),
        ugc.UGC("MD", "Z", "006"),
        ugc.UGC("MD", "Z", "007"),
        ugc.UGC("MD", "Z", "009"),
        ugc.UGC("MD", "Z", "010"),
        ugc.UGC("MD", "Z", "011"),
        ugc.UGC("MD", "Z", "013"),
        ugc.UGC("MD", "Z", "014"),
        ugc.UGC("MD", "Z", "016"),
        ugc.UGC("MD", "Z", "017"),
        ugc.UGC("MD", "Z", "018"),
        ugc.UGC("VA", "Z", "036"),
        ugc.UGC("VA", "Z", "037"),
        ugc.UGC("VA", "Z", "038"),
        ugc.UGC("VA", "Z", "039"),
        ugc.UGC("VA", "Z", "040"),
        ugc.UGC("VA", "Z", "041"),
        ugc.UGC("VA", "Z", "042"),
        ugc.UGC("VA", "Z", "050"),
        ugc.UGC("VA", "Z", "051"),
        ugc.UGC("VA", "Z", "052"),
        ugc.UGC("VA", "Z", "053"),
        ugc.UGC("VA", "Z", "054"),
        ugc.UGC("VA", "Z", "055"),
        ugc.UGC("VA", "Z", "056"),
        ugc.UGC("VA", "Z", "057"),
    ]

    assert ugcs == ugcs_answer
    assert expire == expire_answer



"""Parsing NESDIS Satellite Cloud Product (SCP)."""
import re
from datetime import timedelta
from collections import namedtuple

from pyiem.nws.product import TextProduct

SCP = namedtuple(
    "SCP", ["station", "valid", "mid", "high", "cldtop1", "cldtop2", "eca"]
)
LINEFMT = re.compile(r"^([A-Z0-9]){3,5}\s+[0-3][0-9]/[0-2][0-9][0-5][0-9]\s")


def _to_str(val):
    """Safe conversion."""
    if val.strip() == "":
        return None
    return val.strip()


def _to_int(val, multi=100.0):
    """Safe conversion."""
    if val.strip() == "":
        return None
    try:
        return int(val) * multi
    except ValueError:
        return None


def _processor(textprod):
    """Parse out what we can find in the text product."""
    res = []
    for line in textprod.unixtext.split("\n"):
        if not LINEFMT.match(line):
            continue
        if line[9] != "/":
            continue
        station = line[:5].strip()
        da = int(line[7:9])
        hr = int(line[10:12])
        mi = int(line[12:14])
        valid = textprod.valid
        if da != textprod.valid.day:
            # Yesterday
            valid = textprod.valid - timedelta(days=1)
        valid = valid.replace(day=da, hour=hr, minute=mi)
        mid = _to_str(line[17:20])
        high = _to_str(line[23:26])
        cldtop1 = _to_int(line[28:31])
        cldtop2 = _to_int(line[32:35])
        eca = _to_int(line[37:40], 1)
        res.append(
            SCP._make([station, valid, mid, high, cldtop1, cldtop2, eca])
        )

    return res


class SCPProduct(TextProduct):
    """Class representing a SCP Product"""

    def __init__(self, text, utcnow=None):
        """Constructor

        Args:
          text (str): text to parse
        """
        TextProduct.__init__(self, text, utcnow=utcnow)
        self.data = _processor(self)

    def sql(self, txn):
        """Do the necessary database work

        Args:
          (psycopg2.transaction): a database transaction
        """
        inserts = 0
        for ob in self.data:
            inserts += 1
            txn.execute(
                "INSERT into scp_alldata(station, valid, mid, high, cldtop1, "
                "cldtop2, eca, source) "
                "VALUES (%s, %s, %s, %s, %s, %s, %s, %s)",
                (
                    ob.station,
                    ob.valid,
                    ob.mid,
                    ob.high,
                    ob.cldtop1,
                    ob.cldtop2,
                    ob.eca,
                    self.afos[-1],
                ),
            )
        return inserts


def parser(text, utcnow=None):
    """parser of raw SCP Text.

    Args:
      text (str): the raw text to parse
      utcnow (datetime): the current datetime with timezone set!

    Returns:
      SCPProduct instance
    """
    return SCPProduct(text, utcnow=utcnow)



"""Testing FFG parsing."""

import psycopg2.extras
import pytest
from pyiem.nws.products.ffg import parser as ffgparser
from pyiem.util import get_dbconn, get_test_file


@pytest.fixture
def dbcursor():
    """Return a database cursor."""
    return get_dbconn("postgis").cursor(
        cursor_factory=psycopg2.extras.DictCursor
    )


def test_ffg(dbcursor):
    """FFG"""
    prod = ffgparser(get_test_file("FFGJAN.txt"))
    prod.sql(dbcursor)
    assert len(prod.data.index) == 53


def test_ffg2(dbcursor):
    """FFGKY"""
    prod = ffgparser(get_test_file("FFGKY.txt"))
    prod.sql(dbcursor)
    assert len(prod.data.index) == 113


def test_ffgama(dbcursor):
    """FFGAMA"""
    prod = ffgparser(get_test_file("FFGAMA.txt"))
    prod.sql(dbcursor)
    assert len(prod.data.index) == 23



"""The Atomic Local Storm Report ... Report"""
# pylint: disable=unsubscriptable-object
import re
from datetime import timezone, timedelta
import warnings

from pyiem import reference
from pyiem.util import html_escape

MAG_UNITS = re.compile(
    r"(ACRE|INCHES|INCH|MILE|MPH|KTS|U|FT|F|E|M|TRACE)", re.IGNORECASE
)
# Products that are considered delayed reports
DELAYED_THRESHOLD = timedelta(hours=12)


def _generate_channels(lsrobj):
    """Generate the channels string for this LSR."""
    cleantype = lsrobj.typetext.replace(" ", "_")
    res = [
        f"LSR{lsrobj.wfo}",
        "LSR.ALL",
        f"LSR.{cleantype}",
        f"LSR.{lsrobj.state}",
        f"LSR.{lsrobj.state}.{cleantype}",
    ]
    return ",".join(res)


def _mylowercase(text):
    """ Specialized lowercase function """
    tokens = text.split()
    for i, t in enumerate(tokens):
        if len(t) > 3:
            tokens[i] = t.title()
        elif t in [
            "N",
            "NNE",
            "NNW",
            "NE",
            "E",
            "ENE",
            "ESE",
            "SE",
            "S",
            "SSE",
            "SSW",
            "SW",
            "W",
            "WSW",
            "WNW",
            "NW",
        ]:
            continue
    return " ".join(tokens)


class LSR:
    """ Represents a single Local Storm Report within the LSRProduct """

    def __init__(self):
        """ constructor """
        self.utcvalid = None
        self.valid = None
        self.typetext = None
        self.geometry = None
        self.city = None
        self.county = None
        self.source = None
        self.remark = None
        self.magnitude_f = None
        self.magnitude_str = None
        self.magnitude_qualifier = None
        self.magnitude_units = None
        self.state = None
        self.source = None
        self.text = None
        self.wfo = None
        self.duplicate = False
        self.z = None
        # Carry a reference to the product that had this LSR
        self.product = None

    def __str__(self):
        """String Representation."""
        s = ""
        for attr in self.__dict__:
            s += "%s %s\n" % (attr, getattr(self, attr, ""))
        return s

    def get_lat(self):
        """Return the LSR latitude."""
        return self.geometry.xy[1][0]

    def get_lon(self):
        """Return the LSR longitude."""
        return self.geometry.xy[0][0]

    def consume_magnitude(self, text):
        """ Convert LSR magnitude text into something atomic """
        self.magnitude_str = text
        tokens = MAG_UNITS.findall(text)
        if not tokens:
            if text != "":
                self.product.warnings.append(f"Unable to parse Units |{text}|")
            return
        if len(tokens) == 2:
            self.magnitude_qualifier = tokens[0]
            self.magnitude_units = tokens[1]
        elif len(tokens) == 1:
            self.magnitude_units = tokens[0]
        val = MAG_UNITS.sub("", text).strip()
        if val != "":
            self.magnitude_f = float(val)

    def get_dbtype(self):
        """ Return the typecode used in the database for this event type """
        return reference.lsr_events.get(self.typetext.upper(), None)

    def sql(self, txn):
        """ Provided a database transaction object, persist this LSR """
        wkt = f"SRID=4326;{self.geometry.wkt}"
        # Newer schema supports range partitioning, so can direct insert
        sql = (
            "INSERT into lsrs (valid, type, magnitude, city, county, "
            "state, source, remark, geom, wfo, typetext, product_id, updated, "
            "unit, qualifier) values "
            "(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)"
        )
        args = (
            self.utcvalid,
            self.get_dbtype(),
            self.magnitude_f,
            self.city,
            self.county,
            self.state,
            self.source,
            self.remark,
            wkt,
            self.wfo,
            self.typetext,
            self.product.get_product_id(),
            self.product.valid,
            self.magnitude_units,
            self.magnitude_qualifier,
        )
        txn.execute(sql, args)

    def get_jabbers(self, uri):
        """Return a Jabber formatted message tuple."""
        url = "%s#%s/%s/%s" % (
            uri,
            self.wfo,
            self.utcvalid.strftime("%Y%m%d%H%M"),
            self.utcvalid.strftime("%Y%m%d%H%M"),
        )
        time_fmt = "%-I:%M %p"
        # Is this a delayed report?
        if (self.product.valid - self.valid) > DELAYED_THRESHOLD:
            time_fmt = "%-d %b, %-I:%M %p"
        if self.valid.day != self.product.utcnow.day:
            time_fmt = "%-d %b, %-I:%M %p"

        prefix = ""
        timefmt = "At %-I:%M %p"
        # Is this product delayed?
        if (self.product.valid - self.valid) > DELAYED_THRESHOLD:
            prefix = "[Delayed Report] "
            timefmt = "On %b %-d, at %-I:%M %p"
        magstr = self.mag_string()
        tweet = ("%s%s %s, %s [%s Co, %s] %s reports %s") % (
            prefix,
            self.valid.strftime(timefmt),
            self.z,
            _mylowercase(self.city),
            self.county.title(),
            self.state,
            self.source,
            magstr,
        )
        remainsize = reference.TWEET_CHARS - 24 - len(tweet)
        remark = ""
        if self.remark is not None:
            remark = self.remark.replace("DELAYED REPORT.", "")
            extra = "..." if len(remark) > (remainsize - 6) else ""
            tweet = "%s. %s%s" % (
                tweet,
                remark[: (remainsize - 6)].strip(),
                extra,
            )
        # rectify
        tweet = " ".join(tweet.split())

        xtra = dict(
            product_id=self.product.get_product_id(),
            channels=_generate_channels(self),
            geometry="POINT(%s %s)" % (self.get_lon(), self.get_lat()),
            ptype=self.get_dbtype(),
            valid=self.utcvalid.strftime("%Y%m%dT%H:%M:00"),
            category="LSR",
            twitter="%s %s" % (tweet, url),
            lat=str(self.get_lat()),
            long=str(self.get_lon()),
        )
        html = (
            '<p>%s%s [%s Co, %s] %s <a href="%s">reports %s</a> at '
            "%s %s -- %s</p>"
        ) % (
            prefix,
            _mylowercase(self.city),
            self.county.title(),
            self.state,
            self.source,
            url,
            magstr,
            self.valid.strftime(time_fmt),
            self.z,
            html_escape(remark),
        )

        plain = "%s%s [%s Co, %s] %s reports %s at %s %s -- %s %s" % (
            prefix,
            _mylowercase(self.city),
            self.county.title(),
            self.state,
            self.source,
            magstr,
            self.valid.strftime(time_fmt),
            self.z,
            html_escape(remark),
            url,
        )
        return [plain, html, xtra]

    def tweet(self):
        """return a tweet text"""
        warnings.warn(
            "tweet() is depreciated, use get_jabbers(uri) instead",
            DeprecationWarning,
        )
        j = self.get_jabbers("")
        return j[2]["twitter"]

    def assign_timezone(self, tz, z):
        """ retroactive assignment of timezone, so to improve attrs """
        # We can't just assign the timezone (maybe we can someday)
        self.utcvalid = self.valid + timedelta(hours=reference.offsets[z])
        self.utcvalid = self.utcvalid.replace(tzinfo=timezone.utc)
        self.valid = self.utcvalid.astimezone(tz)
        # complexity with non-DST sites
        if z.endswith("ST") and self.valid.dst():
            self.valid -= timedelta(hours=1)
        self.z = z

    def mag_string(self):
        """ Return a string representing the magnitude and units """
        mag_long = str(self.typetext)
        if self.magnitude_units == "MPH":
            mag_long = "%s of %s%.0f %s" % (
                mag_long,
                self.magnitude_qualifier,
                self.magnitude_f,
                self.magnitude_units,
            )
        elif (
            self.typetext == "HAIL"
            and self.magnitude_f is not None
            and ("%.2f" % (self.magnitude_f,)) in reference.hailsize
        ):
            haildesc = reference.hailsize["%.2f" % (self.magnitude_f,)]
            mag_long = "%s of %s size (%s%.2f %s)" % (
                mag_long,
                haildesc,
                self.magnitude_qualifier,
                self.magnitude_f,
                self.magnitude_units,
            )
        elif self.magnitude_units == "F":
            # Report Tornados as EF scale and not F
            mag_long = "%s of E%s" % (mag_long, self.magnitude_str)
        elif self.magnitude_f:
            mag_long = "%s of %.2f %s" % (
                mag_long,
                self.magnitude_f,
                self.magnitude_units,
            )
        elif self.magnitude_str:
            mag_long = "%s of %s" % (mag_long, self.magnitude_str)
        return mag_long



"""Parser of the Daily Summary Message (DSM)."""
import re
from datetime import datetime, timedelta

from metpy.units import units
from pyiem.nws.product import TextProduct
from pyiem.util import utc
from pyiem.reference import TRACE_VALUE


PARSER_RE = re.compile(
    r"""^(?P<station>[A-Z][A-Z0-9]{3})\s+
   DS\s+
   (COR\s)?
   ([0-9]{4}\s)?
   (?P<day>\d\d)/(?P<month>\d\d)\s?
   ((?P<highmiss>M)|((?P<high>(-?\d+))(?P<hightime>[0-9]{4})))/\s?
   ((?P<lowmiss>M)|((?P<low>(-?\d+))(?P<lowtime>[0-9]{4})))//\s?
   (?P<coophigh>(-?\d+|M))/\s?
   (?P<cooplow>(-?\d+|M))//
   (?P<minslp>M|[\-0-9]{3,4})(?P<slptime>[0-9]{4})?/
   (?P<pday>T|M|[0-9]{,4})/
    (?P<p01>T|M|\-|\-?[0-9]{,4})/(?P<p02>T|M|\-|\-?[0-9]{,4})/
    (?P<p03>T|M|\-|\-?[0-9]{,4})/(?P<p04>T|M|\-|\-?[0-9]{,4})/
    (?P<p05>T|M|\-|\-?[0-9]{,4})/(?P<p06>T|M|\-|\-?[0-9]{,4})/
    (?P<p07>T|M|\-|\-?[0-9]{,4})/(?P<p08>T|M|\-|\-?[0-9]{,4})/
    (?P<p09>T|M|\-|\-?[0-9]{,4})/(?P<p10>T|M|\-|\-?[0-9]{,4})/
    (?P<p11>T|M|\-|\-?[0-9]{,4})/(?P<p12>T|M|\-|\-?[0-9]{,4})/
    (?P<p13>T|M|\-|\-?[0-9]{,4})/(?P<p14>T|M|\-|\-?[0-9]{,4})/
    (?P<p15>T|M|\-|[\-?0-9]{,4})/(?P<p16>T|M|\-|\-?[0-9]{,4})/
    (?P<p17>T|M|\-|\-?[0-9]{,4})/(?P<p18>T|M|\-|\-?[0-9]{,4})/
    (?P<p19>T|M|\-|\-?[0-9]{,4})/(?P<p20>T|M|\-|\-?[0-9]{,4})/
    (?P<p21>T|M|\-|\-?[0-9]{,4})/(?P<p22>T|M|\-|\-?[0-9]{,4})/
    (?P<p23>T|M|\-|\-?[0-9]{,4})/(?P<p24>T|M|\-|\-?[0-9]{,4})/
   (?P<avg_sped>M|\-|[0-9]{2,3})/
   ((?P<drct_sped_max>[0-9]{2})
    (?P<sped_max>[0-9]{2,3})(?P<time_sped_max>[0-9]{4})/
    (?P<drct_gust_max>[0-9]{2})
    (?P<sped_gust_max>[0-9]{2,3})(?P<time_sped_gust_max>[0-9]{4}))?
""",
    re.VERBOSE,
)


def process(text):
    """Emit DSMProduct object for what we can parse."""
    m = PARSER_RE.match(text.replace("\r\r\n", "").replace("\n", ""))
    if m is None:
        return None
    dsm = DSMProduct(m.groupdict())
    return dsm


def compute_time(date, timestamp):
    """Make a valid timestamp."""
    if timestamp is None:
        return None
    return datetime(
        date.year,
        date.month,
        date.day,
        int(timestamp[:2]),
        int(timestamp[2:4]),
    )


class DSMProduct:
    """Represents a single DSM."""

    def __init__(self, groupdict):
        """Contructor."""
        self.date = None
        self.high_time = None
        self.low_time = None
        self.time_sped_max = None
        self.time_sped_gust_max = None
        self.station = groupdict["station"]
        self.groupdict = groupdict

    def tzlocalize(self, tzinfo):
        """Localize the timestamps, tricky."""
        offset = tzinfo.utcoffset(datetime(2000, 1, 1)).total_seconds()
        for name in [
            "high_time",
            "low_time",
            "time_sped_max",
            "time_sped_gust_max",
        ]:
            val = getattr(self, name)
            if val is None:
                continue
            # Need to convert timestamp into standard time time, tricky
            ts = val - timedelta(seconds=offset)
            setattr(
                self,
                name,
                utc(ts.year, ts.month, ts.day, ts.hour, ts.minute).astimezone(
                    tzinfo
                ),
            )

    def compute_times(self, utcnow):
        """Figure out when this DSM is valid for."""
        if utcnow is None:
            utcnow = utc()
        ts = utcnow.replace(
            day=int(self.groupdict["day"]), month=int(self.groupdict["month"])
        )
        # Is this ob from 'last year'
        if ts.month == 12 and utcnow.month == 1:
            ts = ts.replace(year=(ts.year - 1))
        self.date = datetime(ts.year, ts.month, ts.day).date()
        self.high_time = compute_time(
            self.date, self.groupdict.get("hightime")
        )
        self.low_time = compute_time(self.date, self.groupdict.get("lowtime"))
        self.time_sped_max = compute_time(
            self.date, self.groupdict.get("time_sped_max")
        )
        self.time_sped_gust_max = compute_time(
            self.date, self.groupdict.get("time_sped_gust_max")
        )

    def sql(self, txn):
        """Persist to database given the transaction object."""
        cols = []
        args = []

        val = self.groupdict.get("high")
        if val is not None and val != "M":
            cols.append("max_tmpf")
            args.append(val)

        val = self.groupdict.get("low")
        if val is not None and val != "M":
            cols.append("min_tmpf")
            args.append(val)

        val = self.groupdict.get("pday")
        if val is not None and val != "M":
            cols.append("pday")
            args.append(TRACE_VALUE if val == "T" else float(val) / 100.0)

        val = self.groupdict.get("sped_max")
        if val is not None:
            cols.append("max_sknt")
            args.append(
                (int(val) * units("miles / hour")).to(units("knots")).m
            )

        val = self.time_sped_max
        if val is not None:
            cols.append("max_sknt_ts")
            args.append(val)

        val = self.groupdict.get("sped_gust_max")
        if val is not None:
            cols.append("max_gust")
            args.append(
                (int(val) * units("miles / hour")).to(units("knots")).m
            )

        val = self.time_sped_gust_max
        if val is not None:
            cols.append("max_gust_ts")
            args.append(val)

        if not cols:
            return False
        cs = ", ".join(["%s = %%s" % (c,) for c in cols])
        slicer = slice(0, 4) if self.station[0] != "K" else slice(1, 4)
        args.extend([self.station[slicer], self.date])
        txn.execute(
            (
                f"UPDATE summary_{self.date.year} s SET {cs} FROM stations t "
                "WHERE s.iemid = t.iemid and t.network ~* 'ASOS' "
                "and t.id = %s and s.day = %s"
            ),
            args,
        )
        return txn.rowcount == 1


class DSMCollective(TextProduct):
    """A collective representing a NOAAPort Text Product with many DSMs."""

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)
        # hold our parsing results
        self.data = []
        lines = self.unixtext.split("\n")
        if len(lines[3]) < 10:
            meat = ("".join(lines[4:])).split("=")
        else:
            meat = ("".join(lines[3:])).split("=")
        for piece in meat:
            if piece == "":
                continue
            res = process(piece)
            if res is None:
                self.warnings.append(f"DSM RE Match Failure: '{piece}'")
                continue
            res.compute_times(utcnow)
            self.data.append(res)

    def tzlocalize(self, tzprovider):
        """Localize our currently stored timestamps."""
        for dsm in self.data:
            tzinfo = tzprovider.get(dsm.station)
            if tzinfo is None:
                self.warnings.append(f"station {dsm.station} has no tzinfo")
                continue
            dsm.tzlocalize(tzinfo)

    def sql(self, txn):
        """Do databasing."""
        res = []
        for dsm in self.data:
            res.append(dsm.sql(txn))
        return res


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Provide back DSM objects based on the parsing of this text """
    return DSMCollective(text, utcnow, ugc_provider, nwsli_provider)



"""Can we make windrose plots?"""

import pytest
import numpy as np
from metpy.units import units
from pyiem.plot.windrose import WindrosePlot, histogram, plot

BINS = np.array([2, 5, 10, 20]) * units("mph")
SPEED = np.arange(0, 60, 0.1) * units("meter / second")
DIRECTION = SPEED.m * 6.0 * units("degree")


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_windrose_basic():
    """Can we crawl."""
    wr = WindrosePlot()
    return wr.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_windrose_barplot_noopts():
    """Generate a plot with defaults for bins."""
    wr = plot(DIRECTION, SPEED, rmax=10)
    return wr.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_windrose_barplot():
    """Can we crawl."""
    wr = plot(DIRECTION, SPEED, bins=BINS, nsector=16)
    return wr.fig


def test_histogram():
    """Can we make histograms properly."""
    nsector = 16
    calm_percent, dirbins, table = histogram(SPEED, DIRECTION, BINS, nsector)
    # 9 out of 600 above are below 2 MPH, 1.5%
    assert abs(calm_percent.m - 1.50) < 0.01
    assert dirbins.m.shape[0] == table.m.shape[0]



"""IEM Tracker Related Stuff."""
import datetime
import smtplib
from email.mime.text import MIMEText

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from pyiem.util import get_dbconn


class TrackerEngine:
    """A processing engine of tracking offline/online events"""

    def __init__(self, icursor, pcursor, maxactions=0):
        """Constructor of TrackerEngine object

        We need to be provided with psycopg2 cursors to both the `iem` database
        and the `portfolio` database as we have business logic in both places

        Args:
          icursor (cursor): psycopg2 cursor to the iem database
          pcursor (cursor): psycopg2 cursor to the portfolio database
          maxactions (int, optional): threshold for now many email actions we
            allow before we don't wish to spam our users.  0 implies no limit

        """
        self.icursor = icursor
        self.pcursor = pcursor
        self.maxactions = maxactions
        self.action_count = 0
        self.emails = {}

    def send_emails(self, really_send=True):
        """Send out those SPAM emails!"""
        # Don't do anything if we have exceeded maxoffline
        if self.action_count >= self.maxactions and self.maxactions > 0:
            return
        if not really_send:
            return
        s = smtplib.SMTP()
        s.connect()
        for email in self.emails:
            msg = MIMEText(self.emails[email]["body"])
            msg["From"] = "akrherz@iastate.edu"
            msg["Subject"] = self.emails[email]["subject"]
            s.sendmail(msg["From"], email, msg.as_string())
        s.close()

    def offline_logic(self, sid, ob, pnetwork, nt):
        """offline logic

        Args:
          sid (str): site identifier
          ob (dict): observation dictionary
          pnetwork (str): Portfolio name of this network
          nt (dict): provider of station metadata

        """
        # Get a listing of OPEN tickets
        open_tickets = ""
        self.pcursor.execute(
            "SELECT id, entered, subject from tt_base WHERE portfolio = %s "
            "and s_mid = %s and status != 'CLOSED' ORDER by id DESC",
            (pnetwork, sid),
        )
        for row in self.pcursor:
            open_tickets += (" %-6s %16s     %s\n" "") % (
                row[0],
                row[1].strftime("%Y-%m-%d %I %p"),
                row[2],
            )
        # Get a listing of past 4 closed tickets
        closed_tickets = ""
        self.pcursor.execute(
            "SELECT id, entered, subject from tt_base WHERE portfolio = %s "
            "and s_mid = %s and status = 'CLOSED' ORDER by id DESC LIMIT 5",
            (pnetwork, sid),
        )
        for row in self.pcursor:
            closed_tickets += (" %-6s %16s     %s\n" "") % (
                row[0],
                row[1].strftime("%Y-%m-%d %I %p"),
                row[2],
            )
        if closed_tickets == "":
            closed_tickets = " --None-- "
        if open_tickets == "":
            open_tickets = " --None-- "
        # Create an entry in tt_base
        self.pcursor.execute(
            "INSERT into tt_base (portfolio, s_mid, subject, "
            "status, author) VALUES (%s, %s, %s, %s, %s) RETURNING id",
            (pnetwork, sid, "Site Offline", "OPEN", "mesonet"),
        )
        trackerid = self.pcursor.fetchone()[0]
        # Create a tt_log entry
        lts = ob["valid"].astimezone(ZoneInfo(nt.sts[sid]["tzname"]))
        msg = "Site Offline since %s" % (lts.strftime("%d %b %Y %H:%M %Z"),)
        self.pcursor.execute(
            "INSERT into tt_log (portfolio, s_mid, author, status_c, "
            "comments, tt_id) VALUES (%s, %s, %s, %s, %s, %s)",
            (pnetwork, sid, "mesonet", "OKAY", msg, trackerid),
        )

        # Update iemaccess
        self.icursor.execute(
            "INSERT into offline(station, network, "
            "valid, trackerid) VALUES (%s, %s, %s, %s)",
            (sid, nt.sts[sid]["network"], ob["valid"], trackerid),
        )

        mailstr = f"""
----------------------
| IEM TRACKER REPORT |  New Ticket Generated: # {trackerid}
================================================================
 ID                :  {sid} [IEM Network: {nt.sts[sid]["network"]}]
 Station Name      :  {nt.sts[sid]["name"]}
 Status Change     :  [OFFLINE] Site is NOT reporting to the IEM
 Last Observation  :  {lts.strftime("%d %b %Y %I:%M %p %Z")}

 Other Currently 'Open' Tickets for this Site:
 #      OPENED_ON            TICKET TITLE
{open_tickets}

 Most Recently 'Closed' Trouble Tickets for this Site:
 #      CLOSED_ON            TICKET TITLE
{closed_tickets}
================================================================
"""
        # Get contacts for site
        self.pcursor.execute(
            "SELECT distinct email from iem_site_contacts "
            "WHERE s_mid = %s and email is not NULL",
            (sid,),
        )
        for row in self.pcursor:
            email = row[0].lower()
            if email not in self.emails:
                subject = f"[IEM] {nt.sts[sid]['name']} Offline"
                self.emails[email] = {"subject": subject, "body": mailstr}
            else:
                subject = "[IEM] Multiple Sites"
                self.emails[email]["subject"] = subject
                self.emails[email]["body"] += "\n=========\n"
                self.emails[email]["body"] += mailstr

    def online_logic(self, sid, offline, ob, pnetwork, nt):
        """online logic

        Args:
          sid (str): site identifier
          offline (dict): dictionary of offline metadata
          ob (dict): observation dictionary
          pnetwork (str): Portfolio name of this network
          nt (dict): provider of station metadata

        """
        trackerid = offline[sid]["trackerid"]
        # Create Log Entry
        cmt = ("Site Back Online at: %s" "") % (
            ob["valid"].strftime("%Y-%m-%d %H:%M:%S"),
        )
        self.pcursor.execute(
            "INSERT into tt_log (portfolio, s_mid, author, status_c, "
            "comments, tt_id) VALUES (%s, %s, %s, %s, %s, %s)",
            (pnetwork, sid, "mesonet", "CLOSED", cmt, trackerid),
        )
        # Update tt_base
        self.pcursor.execute(
            "UPDATE tt_base SET last = now(), status = 'CLOSED' WHERE id = %s",
            (trackerid,),
        )
        # Update iemaccess
        self.icursor.execute(
            "DELETE from offline where station = %s and network = %s",
            (sid, nt.sts[sid]["network"]),
        )
        ltz = ZoneInfo(nt.sts[sid]["tzname"])
        lts = ob["valid"].astimezone(ltz)
        delta = ob["valid"] - offline[sid]["valid"]
        days = delta.days
        hours = delta.seconds / 3600.0
        minutes = (delta.seconds % 3600) / 60.0
        duration = "%.0f days %.0f hours %.0f minutes" % (days, hours, minutes)
        mailstr = f"""
   ---------------------------------
   |  *** IEM TRACKER REPORT ***   |
------------------------------------------------------------
ID                :  {sid} [IEM Network: {nt.sts[sid]["network"]}]
Station Name      :  {nt.sts[sid]["name"]}
Status Change     :  [ONLINE] Site is reporting to the IEM
Trouble Ticket#   :  {trackerid}

Last Observation  :  {lts.strftime("%d %b %Y %I:%M %p %Z")}
Outage Duration   :  {duration}

IEM Tracker Action:  This trouble ticket has been marked
                     CLOSED pending any further information.
------------------------------------------------------------

  * If you have any information pertaining to this outage,
    please directly respond to this email.
  * Questions about this alert?  Email:  akrherz@iastate.edu
  * Thanks!!!
"""
        # Get contacts for site
        self.pcursor.execute(
            "SELECT distinct email from iem_site_contacts WHERE "
            "s_mid = %s and email is not NULL",
            (sid,),
        )
        for row in self.pcursor:
            email = row[0].lower()
            if email not in self.emails:
                subject = ("[IEM] %s Online" "") % (nt.sts[sid]["name"],)
                self.emails[email] = {"subject": subject, "body": mailstr}
            else:
                subject = "[IEM] Multiple Sites"
                self.emails[email]["subject"] = subject
                self.emails[email]["body"] += "\n=========\n"
                self.emails[email]["body"] += mailstr

    def process_network(self, obs, pnetwork, nt, threshold):
        """Process a list of dicts representing the network's observations

        Args:
          obs (list): list of dicts representing the network obs
          pnetwork (str): the identifier of this network used in Portfolio DB
          nt (NetworkTable): dictionary provider of station metadata
          threshold (datetime): datetime instance with tzinfo set represeting
            the minimum time a site is considered to be 'online' within

        """
        network = nt.sts[list(nt.sts.keys())[0]]["network"]
        self.icursor.execute(
            "SELECT station, trackerid, valid from offline WHERE network = %s",
            (network,),
        )
        offline = {}
        for row in self.icursor:
            offline[row[0]] = {"trackerid": row[1], "valid": row[2]}

        for sid in obs:
            ob = obs[sid]
            if ob["valid"] > threshold:
                # print '%s is online, offlinekeys: %s' % (sid,
                #                                         str(offline.keys()))
                if sid in offline:
                    self.action_count += 1
                    self.online_logic(sid, offline, ob, pnetwork, nt)
                continue
            elif sid in offline:
                # NOOP
                # print '%s is offline and known offline' % (sid, )
                continue
            # We must act!
            # print '%s is offline' % (sid, )
            self.action_count += 1
            self.offline_logic(sid, ob, pnetwork, nt)


def loadqc(cursor=None, date=None):
    """Load the current IEM Tracker QC'd variables

    Args:
      cursor (cursor,optional): Optionally provided database cursor
      date (date,optional): Defaults to today, which tickets are valid for now
    """
    if date is None:
        date = datetime.date.today()
    qdict = {}
    if cursor is None:
        portfolio = get_dbconn("portfolio", user="nobody")
        cursor = portfolio.cursor()

    cursor.execute(
        "select s_mid, sensor, status from tt_base WHERE sensor is not null "
        "and date(entered) <= %s and (status != 'CLOSED' or closed > %s) "
        "and s_mid is not null",
        (date, date),
    )
    for row in cursor:
        sid = row[0]
        if row[0] not in qdict:
            qdict[sid] = {}
        for vname in row[1].split(","):
            qdict[sid][vname.strip()] = True
    return qdict



"""Test HVTEC."""
from pyiem.nws import hvtec


def test_badtime():
    """Make sure contime errors"""
    v = hvtec.contime("BAD")
    assert v is None


def test_empty():
    """ check empty HVTEC Parsing """
    v = hvtec.parse("/00000.0.ER.000000T0000Z.000000T0000Z.000000T0000Z.OO/")
    assert v[0].nwsli.id == "00000"


def test_empty2():
    """Test empty."""
    initial = "/NWYI3.0.ER.000000T0000Z.000000T0000Z.000000T0000Z.OO/"
    v = hvtec.parse(initial)
    assert v[0].nwsli.id == "NWYI3"
    assert str(v[0]) == initial



"""
 Something to deal with SPC PTS Product
 My life was not supposed to end like this, what a brutal format
"""
import re
import datetime
import copy
import os
import itertools

import numpy as np
from shapely.geometry import (
    Polygon,
    LineString,
    MultiPolygon,
    Point,
    MultiPoint,
    MultiLineString,
)
from shapely.geometry.collection import GeometryCollection
from shapely.geometry.polygon import LinearRing
from shapely.ops import split
from shapely.affinity import translate
from pyiem.nws.product import TextProduct
from pyiem.util import utc, LOG

CONUS_BASETIME = utc(2019, 5, 9, 16)
CONUS = {"line": None, "poly": None}
DAYRE = re.compile(
    r"SEVERE WEATHER OUTLOOK POINTS DAY\s+(?P<day>[0-9])", re.IGNORECASE
)
DMATCH = re.compile(r"D(?P<day1>[0-9])\-?(?P<day2>[0-9])?")

THRESHOLD2TEXT = {
    "MRGL": "Marginal",
    "SLGT": "Slight",
    "ENH": "Enhanced",
    "MDT": "Moderate",
    "HIGH": "High",
    "IDRT": "Isolated Dry Thunderstorm",
    "SDRT": "Scattered Dry Thunderstorm",
    "ELEV": "Elevated",
    "CRIT": "Critical",
    "EXTM": "Extreme",
}
THRESHOLD_ORDER = [
    "0.02",
    "0.05",
    "0.10",
    "0.15",
    "0.25",
    "0.30",
    "0.35",
    "0.40",
    "0.45",
    "0.60",
    "TSTM",
    "MRGL",
    "SLGT",
    "ENH",
    "MDT",
    "HIGH",
    "IDRT",
    "SDRT",
    "ELEV",
    "CRIT",
    "EXTM",
]


def compute_times(afos, issue, expire, day):
    """compute actual issue, expire time.

    For the multi-day products, the text product contains a range of dates
    that need translated to an actual issue and expire time.

    Returns:
      issue (datetime)
      expire (datetime)
    """
    if afos not in ["PTSD48", "PFWF38"]:
        return issue, expire
    baseday = 3 if afos == "PFWF38" else 4
    issue = issue + datetime.timedelta(days=(day - baseday))
    return issue, issue + datetime.timedelta(hours=24)


def get_day(text):
    """Figure out which day this is for"""
    search = DAYRE.search(text)
    if search is None:
        return None
    return int(search.groupdict()["day"])


def load_conus_data(valid):
    """ Load up the conus datafile for our perusal """
    fn = "%s/../../data/conus_marine_bnds%s.txt" % (
        os.path.dirname(__file__),
        "_pre190509" if valid < CONUS_BASETIME else "",
    )
    lons = []
    lats = []
    with open(fn) as fh:
        for line in fh:
            tokens = line.split(",")
            lons.append(float(tokens[0]))
            lats.append(float(tokens[1]))
    CONUS["line"] = np.column_stack([lons, lats])
    CONUS["poly"] = Polygon(CONUS["line"])


def get_segments_from_text(text):
    """ Return list of segments for this text """
    tokens = re.findall("([0-9]{8})", text.replace("\n", ""))
    # First we generate a list of segments, based on what we found
    segments = []
    pts = []
    for token in tokens:
        lat = float(token[:4]) / 100.0
        lon = 0 - (float(token[-4:]) / 100.0)
        if lon > -30:
            lon -= 100.0
        if token == "99999999":
            segments.append(pts)
            pts = []
        else:
            pts.append([lon, lat])
    if pts:
        segments.append(pts)

    return segments


def clean_segment(ls):
    """Attempt to get this segment cleaned up.

    Args:
      segment (list): inbound data

    Returns:
      segment (list)
    """

    def _test(val):
        """Our tester."""
        return isinstance(val, MultiPoint) and len(val) == 2

    # If this intersects twice, we are golden
    res = LineString(CONUS["poly"].exterior.coords).intersection(ls)
    if _test(res):
        return ls

    # First and last point of the ls need to be exterior to the CONUS
    for idx in [0, -1]:
        pt = Point(ls.coords[idx])
        if not pt.within(CONUS["poly"]):
            continue
        pt = CONUS["poly"].exterior.interpolate(
            CONUS["poly"].exterior.project(pt)
        )
        if pt.within(CONUS["poly"]):
            LOG.info("     idx: %s is still within, evasive action", idx)
            for xoff, yoff in [
                [-0.01, -0.01],
                [-0.01, 0.0],
                [-0.01, 0.01],
                [0.0, -0.01],
                [0.0, 0.0],
                [0.0, 0.01],
                [0.01, -0.01],
                [0.01, 0.0],
                [0.01, 0.01],
            ]:
                pt2 = translate(pt, xoff=xoff, yoff=yoff)
                if not pt2.within(CONUS["poly"]):
                    pt = pt2
                    LOG.info("     idx: %s is now %s", idx, pt)
                    break
        LOG.info(
            "     fix idx: %s to new: %.4f %.4f Inside: %s",
            idx,
            pt.x,
            pt.y,
            pt.within(CONUS["poly"]),
        )
        coords = list(ls.coords)
        coords[idx] = (pt.x, pt.y)
        ls = LineString(coords)
    res = LineString(CONUS["poly"].exterior.coords).intersection(ls)
    if _test(res):
        return ls

    # Are we doing 3+ intersections already
    if isinstance(res, MultiPoint) and len(res) > 2:
        return MultiLineString(
            [
                r
                for r in ls.intersection(CONUS["poly"])
                if isinstance(r, LineString)
            ]
        )

    LOG.info("     clean_segment failed with res: %s", res)
    return None


def look_for_closed_polygon(segment):
    """Simple logic to see if our polygon is already closed."""
    if segment[0][0] == segment[-1][0] and segment[0][1] == segment[-1][1]:
        LOG.info("Single closed polygon found, done and done")
        return MultiPolygon([Polygon(segment)])

    # Slightly bad line-work, whereby the start and end points are very close
    # to each other
    if (
        (segment[0][0] - segment[-1][0]) ** 2
        + (segment[0][1] - segment[-1][1]) ** 2
    ) ** 0.5 < 0.05:
        LOG.info(
            "assuming linework error, begin: (%.2f %.2f) end: (%.2f %.2f)",
            segment[0][0],
            segment[0][1],
            segment[-1][0],
            segment[-1][1],
        )
        segment[-1] = segment[0]
        return MultiPolygon([Polygon(segment)])


def segment_logic(segment, currentpoly, polys):
    """Our segment parsing logic."""
    if segment[0] == segment[-1] and len(segment) > 2:
        LOG.info("     segment is closed polygon!")
        lr = LinearRing(LineString(segment))
        if not lr.is_ccw:
            LOG.info("     polygon is clockwise (exterior), done.")
            polys.append(currentpoly)
            return Polygon(segment)
        LOG.info("     polygon is CCW (interior), testing intersection")
        if currentpoly.intersection(lr).is_empty:
            LOG.info("     failed intersection with currentpoly, abort")
            return currentpoly
        interiors = [ln for ln in currentpoly.interiors]
        interiors.append(lr)
        newp = Polygon(currentpoly.exterior, interiors)
        if not newp.is_valid:
            LOG.info("     adding interior invalid, buffering")
            newp = newp.buffer(0)
        if newp.is_valid:
            LOG.info(
                "     polygon is interior to currentpoly, area: %.2f ",
                currentpoly.area,
            )
            return newp
        raise Exception(
            "Adding interior polygon resulted in an invalid geometry, aborting"
        )

    # All open lines need to intersect the CONUS, ensure that happens
    ls = LineString(segment)
    ls = clean_segment(ls)
    if isinstance(ls, MultiLineString):
        for _ls in ls:
            LOG.info("     look out below, recursive we go.")
            currentpoly = segment_logic(_ls.coords, currentpoly, polys)
        return currentpoly
    if ls is None:
        LOG.info("     aborting as clean_segment failed...")
        return currentpoly
    LOG.info(
        "     new segment start: %.4f %.4f end: %.4f %.4f",
        ls.coords[0][0],
        ls.coords[0][1],
        ls.coords[-1][0],
        ls.coords[-1][1],
    )

    # If this line segment does not intersect the current polygon of interest,
    # we should check any previous polygons to see if it intersects it. We
    # could be dealing with invalid ordering in the file, sigh.
    if currentpoly.intersection(ls).is_empty:
        LOG.info("     ls does not intersect currentpoly, looking for match")
        found = False
        for i, poly in enumerate(polys):
            intersect = poly.intersection(ls)
            if intersect.is_empty or isinstance(intersect, MultiLineString):
                continue
            LOG.info(
                "     found previous polygon i:%s area: %.1f that intersects",
                i,
                poly.area,
            )
            found = True
            polys.append(currentpoly)
            currentpoly = polys.pop(i)
            break
        if not found:
            LOG.info("     setting currentpoly back to CONUS")
            polys.append(currentpoly)
            currentpoly = copy.deepcopy(CONUS["poly"])

    # Results in either [currentpoly] or [polya, polyb, ...]
    geomcollect = split(currentpoly, ls)
    if len(geomcollect) > 2:
        LOG.info("     line intersects polygon 3+ times, can't handle")
        return currentpoly
    if len(geomcollect) == 1:
        res = geomcollect.geoms[0]
    else:
        (polya, polyb) = geomcollect.geoms[0], geomcollect.geoms[1]
        # Linear reference our splitter's start and end distance
        startdist = polya.exterior.project(Point(ls.coords[0]))
        enddist = polya.exterior.project(Point(ls.coords[-1]))
        # if the end is further down the line, we want this polygon
        res = polya if enddist > startdist else polyb

    if res.area > 0.01:
        LOG.info("     taking polygon.area = %.4f", res.area)
        return res
    return currentpoly


# def debug_draw(segment, current_poly):
#    """Draw this for debugging purposes."""
#    segment = np.array(segment)
#    from pyiem.plot.use_agg import plt
#    (fig, ax) = plt.subplots(1, 1)
#    ax.plot(segment[:, 0], segment[:, 1], c='b')
#    ylim = ax.get_ylim()
#    xlim = ax.get_xlim()
#    ax.plot(current_poly.exterior.xy[0], current_poly.exterior.xy[1], c='r')
#    ax.set_xlim(*xlim)
#    ax.set_ylim(*ylim)
#    LOG.info("writting /tmp/debugdraw.png")
#    fig.savefig("/tmp/debugdraw.png")


def str2multipolygon(s):
    """Convert string PTS data into a polygon.

    Args:
      s (str): the cryptic string that we attempt to make valid polygons from
    """
    segments = get_segments_from_text(s)
    # Simple case whereby the segment is its own circle, thank goodness
    if len(segments) == 1:
        res = look_for_closed_polygon(segments[0])
        if res:
            return res

    # Keep track of generated polygons
    polys = []
    # currentpoly is our present subject of interest
    currentpoly = copy.deepcopy(CONUS["poly"])

    for i, segment in enumerate(segments):
        # debug_draw(segment, currentpoly)
        LOG.info(
            "  Iterate: %s/%s, len(segment): %s (%.2f %.2f) (%.2f %.2f)",
            i + 1,
            len(segments),
            len(segment),
            segment[0][0],
            segment[0][1],
            segment[-1][0],
            segment[-1][1],
        )
        currentpoly = segment_logic(segment, currentpoly, polys)
    polys.append(currentpoly)

    res = []
    LOG.info(
        "  Resulted in len(polys): %s, now quality controlling", len(polys)
    )
    for i, poly in enumerate(polys):
        if not poly.is_valid:
            LOG.info("     ERROR: polygon %s is invalid!", i)
            continue
        if poly.area == CONUS["poly"].area:
            LOG.info("     polygon %s is just CONUS, skipping", i)
            continue
        LOG.info("     polygon: %s has area: %s", i, poly.area)
        res.append(poly)
    if not res:
        raise Exception(
            (
                "Processed no geometries, this is a bug!\n"
                "  s is %s\n"
                "  segments is %s" % (repr(s), repr(segments))
            )
        )
    return MultiPolygon(res)


class SPCOutlookCollection:
    """ A collection of outlooks for a single 'day'"""

    def __init__(self, issue, expire, day):
        """Constructor"""
        self.issue = issue
        self.expire = expire
        self.day = day
        self.outlooks = []


class SPCOutlook:
    """A class holding what we store for a single outlook."""

    def __init__(self, category, threshold, multipoly):
        """Constructor.

        Args:
          category (str): the label of this category
          threshold (str): the threshold associated with the category
          multipoly (MultiPolygon): the geometry
        """
        self.category = category
        self.threshold = threshold
        self.geometry = multipoly
        self.wfos = []


class SPCPTS(TextProduct):
    """A class representing the polygons and metadata in SPC PTS Product"""

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """Constructor

        Args:
          text (string): the raw PTS product that is to be parsed
          utcnow (datetime, optional): in case of ambuigity with time
          ugc_provider (dict, optional): unused in this class
          nwsli_provider (dict, optional): unused in this class
        """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)
        LOG.info("==== SPCPTS Processing: %s", self.get_product_id())
        load_conus_data(self.valid)
        self.issue = None
        self.expire = None
        self.day = None
        self.outlook_type = None
        self.outlook_collections = dict()
        self.set_metadata()
        self.find_issue_expire()
        self.find_outlooks()
        self.quality_control()

    def quality_control(self):
        """Run some checks against what was parsed"""
        # 1. Do polygons overlap for the same outlook
        LOG.info("==== Running Quality Control Checks")
        for day, collect in self.outlook_collections.items():
            # Everything should be smaller than General Thunder, for conv
            tstm = self.get_outlook("CATEGORICAL", "TSTM", day)
            for outlook in collect.outlooks:
                rewrite = False
                # case of single polygon
                if tstm and len(outlook.geometry) == 1:
                    if outlook.geometry.area > tstm.geometry.area:
                        rewrite = True
                        msg = (
                            "Discarding polygon as it is larger than TSTM: "
                            "Day: %s %s %s Area: %.2f TSTM Area: %.2f"
                        ) % (
                            day,
                            outlook.category,
                            outlook.threshold,
                            outlook.geometry.area,
                            tstm.geometry.area,
                        )
                        LOG.info(msg)
                        self.warnings.append(msg)
                # clip polygons to the CONUS
                good_polys = []
                for poly in outlook.geometry:
                    intersect = CONUS["poly"].intersection(poly)
                    if isinstance(intersect, GeometryCollection):
                        for p in intersect:
                            if isinstance(p, Polygon):
                                good_polys.append(p)
                            else:
                                LOG.info("Discarding %s as not polygon", p)
                    else:
                        if isinstance(intersect, Polygon):
                            good_polys.append(intersect)
                        else:
                            LOG.info("Discarding %s as not polygon", intersect)
                outlook.geometry = MultiPolygon(good_polys)

                good_polys = []
                for poly1, poly2 in itertools.permutations(
                    outlook.geometry, 2
                ):
                    if poly1.contains(poly2):
                        rewrite = True
                        msg = (
                            "Discarding overlapping exterior polygon: "
                            "Day: %s %s %s Area: %.2f"
                        ) % (
                            day,
                            outlook.category,
                            outlook.threshold,
                            poly1.area,
                        )
                        LOG.info(msg)
                        self.warnings.append(msg)
                    elif tstm is not None and poly1.area > tstm.geometry.area:
                        rewrite = True
                        msg = (
                            "Discarding polygon as it is larger than TSTM: "
                            "Day: %s %s %s Area: %.2f"
                        ) % (
                            day,
                            outlook.category,
                            outlook.threshold,
                            poly1.area,
                        )
                        LOG.info(msg)
                        self.warnings.append(msg)
                    else:
                        if poly1 not in good_polys:
                            good_polys.append(poly1)
                if rewrite:
                    outlook.geometry = MultiPolygon(good_polys)
        # 2. Do the time bounds make sense, limited scope here
        if (
            self.day == 1
            and (self.issue - self.valid).total_seconds() > 8 * 3600
        ):
            self.warnings.append(
                ("time_bounds_check: day: %s issue: %s valid: %s expire: %s")
                % (self.day, self.issue, self.valid, self.expire)
            )

    def get_outlookcollection(self, day):
        """Returns the SPCOutlookCollection for a given day"""
        return self.outlook_collections.get(day)

    def get_outlook(self, category, threshold, day=None):
        """ Get an outlook by category and threshold """
        if not self.outlook_collections:
            return None
        if day is None:
            day = self.day
        if day not in self.outlook_collections:
            return None
        for outlook in self.outlook_collections[day].outlooks:
            if outlook.category == category and outlook.threshold == threshold:
                return outlook
        return None

    def draw_outlooks(self):
        """ For debugging, draw the outlooks on a simple map for inspection!"""
        from descartes.patch import PolygonPatch
        import matplotlib.pyplot as plt

        for day, collect in self.outlook_collections.items():
            for outlook in collect.outlooks:
                fig = plt.figure(figsize=(12, 8))
                ax = fig.add_subplot(111)
                # pylint: disable=unsubscriptable-object
                ax.plot(
                    CONUS["line"][:, 0],
                    CONUS["line"][:, 1],
                    color="b",
                    label="Conus",
                )
                for poly in outlook.geometry:
                    patch = PolygonPatch(
                        poly,
                        fc="tan",
                        label="Outlook %.1f" % (poly.area,),
                        zorder=2,
                    )
                    ax.add_patch(patch)
                    ax.plot(
                        poly.exterior.xy[0],
                        poly.exterior.xy[1],
                        lw=2,
                        color="r",
                    )
                ax.set_title(
                    ("Day %s Category %s Threshold %s")
                    % (day, outlook.category, outlook.threshold)
                )
                ax.legend(loc=3)
                fn = (
                    ("/tmp/%s_%s_%s_%s.png")
                    % (
                        day,
                        self.issue.strftime("%Y%m%d%H%M"),
                        outlook.category,
                        outlook.threshold,
                    )
                ).replace(" ", "_")
                LOG.info(":: creating plot %s", fn)
                fig.savefig(fn)
                del fig
                del ax

    def set_metadata(self):
        """
        Set some metadata about this product
        """
        if self.afos == "PTSDY1":
            self.day = 1
            self.outlook_type = "C"
        elif self.afos == "PTSDY2":
            self.day = 2
            self.outlook_type = "C"
        elif self.afos == "PTSDY3":
            self.day = 3
            self.outlook_type = "C"
        elif self.afos == "PTSD48":
            self.outlook_type = "C"
        elif self.afos == "PFWFD1":
            self.day = 1
            self.outlook_type = "F"
        elif self.afos == "PFWFD2":
            self.day = 2
            self.outlook_type = "F"
        elif self.afos == "PFWF38":
            self.outlook_type = "F"
        else:
            self.warnings.append(
                ("Unknown awipsid '%s' for metadata") % (self.afos,)
            )

    def find_issue_expire(self):
        """
        Determine the period this product is valid for
        """
        tokens = re.findall("VALID TIME ([0-9]{6})Z - ([0-9]{6})Z", self.text)
        day1 = int(tokens[0][0][:2])
        hour1 = int(tokens[0][0][2:4])
        min1 = int(tokens[0][0][4:])
        day2 = int(tokens[0][1][:2])
        hour2 = int(tokens[0][1][2:4])
        min2 = int(tokens[0][1][4:])
        issue = self.valid.replace(day=day1, hour=hour1, minute=min1)
        expire = self.valid.replace(day=day2, hour=hour2, minute=min2)
        # NB: outlooks can go out more than just one day
        if day1 < self.valid.day:
            issue = self.valid + datetime.timedelta(days=25)
            issue = issue.replace(day=day1, hour=hour1, minute=min1)
        if day2 < self.valid.day:
            expire = self.valid + datetime.timedelta(days=25)
            expire = expire.replace(day=day2, hour=hour2, minute=min2)
        self.issue = issue
        self.expire = expire

    def find_outlooks(self):
        """ Find the outlook sections within the text product! """
        if self.text.find("&&") == -1:
            self.warnings.append("Product contains no &&, adding...")
            self.text = self.text.replace("\n... ", "\n&&\n... ")
            self.text += "\n&& "
        for segment in self.text.split("&&")[:-1]:
            day = self.day
            if day is None:
                day = get_day(segment)
            # We need to figure out the probabilistic or category
            tokens = re.findall(r"\.\.\.\s+(.*)\s+\.\.\.", segment)
            if not tokens:
                continue
            category = tokens[0].strip()
            point_data = {}
            # Now we loop over the lines looking for data
            threshold = None
            for line in segment.split("\n"):
                if (
                    re.match(
                        (
                            r"^(D[3-8]\-?[3-8]?|EXTM|MRGL|ENH|SLGT|MDT|ELEV|"
                            r"HIGH|CRIT|TSTM|SIGN|IDRT|SDRT|0\.[0-9][0-9]) "
                        ),
                        line,
                    )
                    is not None
                ):
                    newthreshold = line.split()[0]
                    if threshold is not None and threshold == newthreshold:
                        point_data[threshold] += " 99999999 "
                    threshold = newthreshold
                if threshold is None:
                    continue
                if threshold not in point_data:
                    point_data[threshold] = ""
                point_data[threshold] += line.replace(threshold, " ")

            if day is not None:
                issue, expire = compute_times(
                    self.afos, self.issue, self.expire, day
                )
                collect = self.outlook_collections.setdefault(
                    day, SPCOutlookCollection(issue, expire, day)
                )
            # We need to duplicate, in the case of day-day spans
            for threshold in list(point_data.keys()):
                if threshold == "TSTM" and self.afos == "PFWF38":
                    LOG.info(("Failing to parse TSTM in PFWF38"))
                    del point_data[threshold]
                    continue
                match = DMATCH.match(threshold)
                if match:
                    data = match.groupdict()
                    if data.get("day2") is not None:
                        day1 = int(data["day1"])
                        day2 = int(data["day2"])
                        LOG.info("Duplicating threshold %s-%s", day1, day2)
                        for i in range(day1, day2 + 1):
                            key = "D%s" % (i,)
                            point_data[key] = point_data[threshold]
                        del point_data[threshold]
            for threshold in point_data:
                match = DMATCH.match(threshold)
                if match:
                    day = int(match.groupdict()["day1"])
                    issue, expire = compute_times(
                        self.afos, self.issue, self.expire, day
                    )
                    collect = self.outlook_collections.setdefault(
                        day, SPCOutlookCollection(issue, expire, day)
                    )
                LOG.info(
                    "--> Start Day: %s Category: '%s' Threshold: '%s' =====",
                    day,
                    category,
                    threshold,
                )
                mp = str2multipolygon(point_data[threshold])
                if DMATCH.match(threshold):
                    threshold = "0.15"
                LOG.info("----> End threshold is: %s", threshold)
                collect.outlooks.append(SPCOutlook(category, threshold, mp))

    def compute_wfos(self, txn):
        """Figure out which WFOs are impacted by this polygon"""
        for day, collect in self.outlook_collections.items():
            for outlook in collect.outlooks:
                if outlook.geometry.is_empty:
                    continue
                sql = """
                    select distinct wfo from ugcs WHERE
                    st_contains(ST_geomFromEWKT('SRID=4326;%s'), centroid) and
                    substr(ugc,3,1) = 'C' and wfo is not null
                    and end_ts is null ORDER by wfo ASC
                """ % (
                    outlook.geometry.wkt,
                )

                txn.execute(sql)
                for row in txn.fetchall():
                    outlook.wfos.append(row["wfo"])
                LOG.info(
                    "Day: %s Category: %s Threshold: %s #WFOS: %s %s",
                    day,
                    outlook.category,
                    outlook.threshold,
                    len(outlook.wfos),
                    ",".join(outlook.wfos),
                )

    def sql(self, txn):
        """Do database work

        Args:
          txn (psycopg2.cursor): database cursor
        """
        for day, collect in self.outlook_collections.items():
            txn.execute(
                """
                DELETE from spc_outlooks where product_issue = %s
                and expire = %s and outlook_type = %s and day = %s
            """,
                (self.valid, self.expire, self.outlook_type, day),
            )
            if txn.rowcount > 0:
                LOG.info(
                    "Removed %s previous spc_outlook entries", txn.rowcount
                )

            for outlook in collect.outlooks:
                if outlook.geometry.is_empty:
                    continue
                sql = """
                    INSERT into spc_outlooks(product_issue, issue, expire,
                    threshold, category, day, outlook_type, geom)
                    VALUES (%s, %s, %s, %s, %s, %s, %s, %s)
                """
                args = (
                    self.valid,
                    collect.issue,
                    collect.expire,
                    outlook.threshold,
                    outlook.category,
                    collect.day,
                    self.outlook_type,
                    "SRID=4326;%s" % (outlook.geometry.wkt,),
                )
                txn.execute(sql, args)

    def get_descript_and_url(self):
        """Helper to convert awips id into strings"""
        product_descript = "((%s))" % (self.afos,)
        url = "http://www.spc.noaa.gov"
        day = "((%s))" % (self.afos,)

        if self.afos == "PTSDY1":
            day = "Day 1"
            product_descript = "Convective"
            url = (
                "https://www.spc.noaa.gov/products/outlook/archive/"
                "%s/day1otlk_%s.html"
            ) % (self.valid.year, self.issue.strftime("%Y%m%d_%H%M"))
        elif self.afos == "PTSDY2":
            day = "Day 2"
            product_descript = "Convective"
            hhmm = "1730" if self.valid.hour > 11 else "0600"
            url = (
                "https://www.spc.noaa.gov/products/outlook/archive/"
                "%s/day2otlk_%s_%s.html"
            ) % (self.valid.year, self.valid.strftime("%Y%m%d"), hhmm)
        elif self.afos == "PTSDY3":
            # 0730 when in CDT, 0830 when in CST
            hhmm = "0730" if self.z == "CDT" else "0830"
            day = "Day 3"
            product_descript = "Convective"
            url = (
                "https://www.spc.noaa.gov/products/outlook/archive/%s/"
                "day3otlk_%s_%s.html"
            ) % (self.valid.year, self.valid.strftime("%Y%m%d"), hhmm)
        elif self.afos == "PTSD48":
            day = "Days 4-8"
            product_descript = "Convective"
            url = (
                "https://www.spc.noaa.gov/products/exper/day4-8/archive/%s/"
                "day4-8_%s.html"
            ) % (self.valid.year, self.valid.strftime("%Y%m%d"))
        elif self.afos == "PFWFD1":
            day = "Day 1"
            product_descript = "Fire Weather"
            url = self.issue.strftime(
                (
                    "https://www.spc.noaa.gov/products/fire_wx/%Y/%y%m%d_%H%M"
                    "_fwdy1_print.html"
                )
            )
        elif self.afos == "PFWFD2":
            day = "Day 2"
            product_descript = "Fire Weather"
            url = self.issue.strftime(
                (
                    "https://www.spc.noaa.gov/products/fire_wx/%Y/%y%m%d_%H%M"
                    "_fwdy2_print.html"
                )
            )
        elif self.afos == "PFWF38":
            day = "Day 3-8"
            product_descript = "Fire Weather"
            url = self.issue.strftime(
                (
                    "https://www.spc.noaa.gov/products/exper/fire_wx/%Y/%y%m%d"
                    ".html"
                )
            )

        return product_descript, url, day

    def get_jabbers(self, uri, _uri2=None):
        """Wordsmith the Jabber/Twitter Messaging

        Examples
        --------
          The Storm Prediction Center issues Day 1 Convective Outlook
            at 16z 6 April
          The Storm Prediction Center issues Day 1 Moderate Risk at 16z 6 Apr
            for portions of DMX
          SPC issues Day 1 Moderate Risk at 16z 6 April for #DMX
        """
        res = []
        product_descript, url, title = self.get_descript_and_url()
        jdict = {
            "title": title,
            "name": "The Storm Prediction Center",
            "tstamp": self.valid.strftime("%b %-d, %-H:%Mz"),
            "outlooktype": product_descript,
            "url": url,
        }
        for _, collect in self.outlook_collections.items():

            wfos = {
                "TSTM": [],
                "EXTM": [],
                "MRGL": [],
                "SLGT": [],
                "ENH": [],
                "CRIT": [],
                "MDT": [],
                "HIGH": [],
                "ELEV": [],
                "IDRT": [],
                "SDRT": [],
            }

            for outlook in collect.outlooks:
                _d = wfos.setdefault(outlook.threshold, [])
                _d.extend(outlook.wfos)

            jdict["day"] = collect.day
            wfomsgs = {}
            # We order in least to greatest, so that the highest threshold
            # overwrites lower ones
            for cat in [
                "MRGL",
                "SLGT",
                "ENH",
                "MDT",
                "HIGH",
                "ELEV",
                "CRIT",
                "EXTM",
                "IDRT",
                "SDRT",
            ]:
                jdict["ttext"] = "%s %s Risk" % (
                    THRESHOLD2TEXT[cat],
                    product_descript,
                )
                for wfo in wfos[cat]:
                    jdict["wfo"] = wfo
                    wfomsgs[wfo] = [
                        (
                            "%(name)s issues Day %(day)s %(ttext)s "
                            "at %(tstamp)s for portions of %(wfo)s %(url)s"
                        )
                        % jdict,
                        (
                            "<p>%(name)s issues "
                            '<a href="%(url)s">Day %(day)s %(ttext)s</a> '
                            "at %(tstamp)s for portions of %(wfo)s's area</p>"
                        )
                        % jdict,
                        {
                            "channels": [
                                wfo,
                                "%s.SPC%s" % (wfo, self.afos[3:]),
                                "%s.SPC%s.%s" % (wfo, self.afos[3:], cat),
                            ],
                            "product_id": self.get_product_id(),
                            "twitter": (
                                "SPC issues Day %(day)s %(ttext)s "
                                "at %(tstamp)s for %(wfo)s %(url)s"
                            )
                            % jdict,
                        },
                    ]
            keys = list(wfomsgs.keys())
            keys.sort()
            res = []
            for wfo in keys:
                res.append(wfomsgs[wfo])

        # Generic for SPC
        res.append(
            [
                (
                    "%(name)s issues %(title)s %(outlooktype)s Outlook at "
                    "%(tstamp)s %(url)s"
                )
                % jdict,
                (
                    '<p>%(name)s issues <a href="%(url)s">%(title)s '
                    "%(outlooktype)s Outlook</a> at %(tstamp)s</p>"
                )
                % jdict,
                {
                    "channels": ["SPC", "SPC%s" % (self.afos[3:],)],
                    "product_id": self.get_product_id(),
                    "twitter": (
                        "%(name)s issues %(title)s "
                        "%(outlooktype)s Outlook at %(tstamp)s %(url)s"
                    )
                    % jdict,
                },
            ]
        )
        return res


def parser(text, utcnow=None):
    """Parse this text!"""
    return SPCPTS(text, utcnow)



"""
http://www.unidata.ucar.edu/data/lightning/nldn.html
"""
from datetime import timezone, datetime, timedelta
import struct

import pandas as pd


class NLDNProduct:
    """Simple class representing obs found in a NLDNProduct."""

    def __init__(self):
        """Constructor."""
        self.header = ""
        self.df = None

    def sql(self, cursor):
        """Persist data to the database"""
        if self.df is None:
            return
        for _, row in self.df.iterrows():
            table = "nldn%s" % (row["valid"].strftime("%Y_%m"),)
            cursor.execute(
                f"INSERT into {table} (valid, geom, signal, multiplicity, "
                "axis, eccentricity, ellipse, chisqr) VALUES (%s, "
                "'SRID=4326;POINT(%s %s)', %s, %s, %s, %s, %s, %s)",
                (
                    row["valid"],
                    row["longitude"],
                    row["latitude"],
                    row["signal"],
                    row["multiplicity"],
                    row["axis"],
                    row["eccentricity"],
                    row["ellipse"],
                    row["chisqr"],
                ),
            )


def parser(buf):
    """NLDN (unencrypted product)"""

    np = NLDNProduct()
    _ = buf.read(4)  # NLDN
    records = struct.unpack(">i", buf.read(4))
    np.header = buf.read(records[0] * 28 - 8)  # SGDS SUNY@ALBANY Thu Sep  ...

    rows = []
    while True:
        chunk = buf.read(28)
        if not chunk:
            break
        (tsec, nsec, lat1000, lon1000) = struct.unpack(">4i", chunk[:16])
        secs = float(tsec) + (nsec / 1000000.0)
        ts = datetime(1970, 1, 1) + timedelta(seconds=secs)
        ts = ts.replace(tzinfo=timezone.utc)
        (_, sgnl10, _) = struct.unpack(">3h", chunk[16:22])
        (multi, _, axis, eccentricity, ellipse, chisqr) = struct.unpack(
            "6b", chunk[22:28]
        )
        rows.append(
            dict(
                valid=ts,
                latitude=lat1000 / 1000.0,
                longitude=lon1000 / 1000.0,
                signal=sgnl10 / 10.0,
                multiplicity=multi,
                axis=axis,
                eccentricity=eccentricity,
                ellipse=ellipse,
                chisqr=chisqr,
            )
        )
    if rows:
        np.df = pd.DataFrame(rows)
    return np



"""Create some data for the scripts to chomp on."""
import datetime

from pyiem.util import get_dbconn, utc


def asos():
    """Create some ASOS data entries."""
    pgconn = get_dbconn("asos")
    cursor = pgconn.cursor()
    basevalid = utc(2015, 1, 1, 6)
    for s in range(360):
        valid = basevalid + datetime.timedelta(hours=s)
        # Keep the max speed at ~24kts
        sknt = s / 13.0
        drct = s
        cursor.execute(
            f"INSERT into t{valid.year} (station, valid, sknt, drct, "
            "report_type) VALUES ('AMW2', %s, %s, %s, 2)",
            (valid, sknt, drct),
        )
    cursor.close()
    pgconn.commit()


def main():
    """Go Main Go."""
    asos()


if __name__ == "__main__":
    main()



"""Test our caching zonal_stats"""

import numpy as np
from geopandas import GeoSeries
from shapely.geometry import Polygon
from affine import Affine
from pyiem.grid import zs


def test_gen_stats():
    """Run a test"""
    affine = Affine(10.0, 0.0, 0.0, 0.0, -10, 100)
    grid = np.reshape(np.arange(100), (10, 10))
    sq1 = Polygon([(50, 50), (50, 60), (60, 60), (60, 50)])
    sq2 = Polygon([(60, 60), (60, 70), (70, 70), (70, 60)])
    # Some polys that are partially off grid
    sq3 = Polygon([(90, 90), (90, 110), (110, 110), (110, 90)])
    sq4 = Polygon([(90, -10), (90, 10), (110, 10), (90, -10)])
    sq5 = Polygon([(-10, -10), (-10, 10), (10, 10), (10, -10)])
    sq6 = Polygon([(-10, 90), (-10, 110), (10, 110), (10, 90)])
    # completely off
    sq7 = Polygon([(-10, -10), (-10, -1), (-1, -1), (-1, -10)])
    geometries = GeoSeries([sq1, sq2, sq3, sq4, sq5, sq6, sq7])
    czs = zs.CachingZonalStats(affine)
    res = czs.gen_stats(np.flipud(grid), geometries)
    assert len(res) == 7
    assert abs(res[0] - 55.0) < 0.1
    assert abs(res[1] - 66.0) < 0.1

    czs = zs.CachingZonalStats(affine)
    res = czs.gen_stats(np.flipud(grid))
    assert not res



"""tests"""
from pyiem import nwnformat


def test_heatindex():
    """Exercise the heatindex func"""
    d = nwnformat.heatidx(100, 50)
    assert abs(d - 119.48) < 0.01
    d = nwnformat.heatidx(69, 50)
    assert abs(d - 69.00) < 0.01
    d = nwnformat.heatidx(169, 50)
    assert abs(d - -99.00) < 0.01
    d = nwnformat.heatidx(79, 0)
    assert abs(d - -99.00) < 0.01
    d = nwnformat.heatidx(75, 40)
    assert abs(d - 73.16) < 0.01
    d = nwnformat.wchtidx(15, 1)
    assert abs(d - 15.00) < 0.01


def test_feels():
    """test the output of the feelslike() method"""
    d = nwnformat.feelslike(100, 50, 10)
    assert abs(d - 119.485) < 0.001
    d = nwnformat.feelslike(10, 50, 10)
    assert abs(d - -3.54) < 0.01


def test_mydir():
    """test the output of the mydir() method"""
    d = nwnformat.mydir(10, 10)
    assert d == 225
    d = nwnformat.mydir(-10, 10)
    assert d == 135
    d = nwnformat.mydir(-10, -10)
    assert d == 45
    d = nwnformat.mydir(10, -10)
    assert d == 315
    d = nwnformat.mydir(10, 0)
    assert d == 270


def test_dwpf():
    """test the output of the dwpf() method"""
    d = nwnformat.dwpf(50, 50)
    assert d == 32
    assert nwnformat.dwpf(None, 32) is None


def test_uv():
    """test the output of the uv() method"""
    u, v = nwnformat.uv(10, 180)
    assert u == 0
    assert v == 10


def test_basic():
    """ basic test of constructor """
    n = nwnformat.nwnformat()
    n.sid = 100
    n.parseLineRT(
        (
            "A 263  14:58 07/16/15   S 09MPH 000K 460F 460F "
            '100% 29.66F 00.00"D 00.00"M 00.00"R'
        ).split()
    )
    n.parseMaxLineRT(
        (
            "A 263    Max 07/16/15   S 21MPH 000K 460F 460F "
            '100% 29.81" 00.00"D 00.00"M 00.00"R'
        ).split()
    )
    n.parseMaxLineRT(
        (
            "A 263    Min 07/16/15   S 01MPH 000K 460F 882F "
            '100% 29.65" 00.00"D 00.00"M 00.00"R'
        ).split()
    )
    n.parseLineRT(
        (
            "A 263  14:59 07/16/15   S 19MPH 000K 460F 460F "
            '100% 29.66F 00.00"D 00.00"M 00.00"R'
        ).split()
    )
    # n.setTS("BAH")
    # self.assertEqual(n.error, 100)
    n.setTS("07/16/15 14:58:50")
    n.sanityCheck()
    n.avgWinds()
    n.currentLine()
    n.maxLine()
    n.minLine()
    assert abs(n.pres - 29.66) < 0.01



"""Process HVTEC encoding."""
import re
from datetime import timezone, datetime

from pyiem.util import LOG
from pyiem.nws.nwsli import NWSLI

#         nwsli        sev         cause
_re = (
    r"(/([A-Z0-9]{5})\.([N0123U])\.([A-Z]{2})\.([0-9TZ]+)\."
    r"([0-9TZ]+)\.([0-9TZ]+)\.([A-Z]{2})/)"
)

_statusDict = {
    "00": "is not applicable",
    "NO": "is not expected",
    "NR": "may be expected",
    "UU": "is not available",
}

_causeDict = {
    "ER": "Excessive Rainfall",
    "SM": "Snowmelt",
    "RS": "Rain and Snowmelt",
    "DM": "Dam or Levee Failure",
    "GO": "Glacier-Dammed Lake Outburst",
    "IJ": "Ice Jam",
    "IC": "Rain and/or Snowmelt and/or Ice Jam",
    "FS": "Upstream Flooding plus Storm Surge",
    "FT": "Upstream Flooding plus Tidal Effects",
    "ET": "Elevated Upstream Flow plus Tidal Effects",
    "WT": "Wind and/or Tidal Effects",
    "DR": "Upstream Dam or Reservoir Release",
    "MC": "Other Multiple Causes",
    "OT": "Other Effects",
    "UU": "Unknown",
}

_severityDict = {
    "N": "None",
    "0": "None",
    "1": "Minor",
    "2": "Moderate",
    "3": "Major",
    "U": "Unknown",
}


def parse(text, nwsli_provider=None):
    """ I look for and return hvtec objects as I find them """
    hvtec = []
    tokens = re.findall(_re, text)
    for t in tokens:
        hvtec.append(HVTEC(t, nwsli_provider))
    return hvtec


def contime(s):
    """Process a string into time."""
    if len(re.findall("0000*T", s)) > 0:
        return None
    try:
        ts = datetime.strptime(s, "%y%m%dT%H%MZ")
        return ts.replace(tzinfo=timezone.utc)
    except Exception as err:
        LOG.exception(err)
        return None


class HVTEC:
    """Hydro VTEC."""

    def __init__(self, tokens, nwsli_provider=None):
        """ Constructor """
        if nwsli_provider is None:
            nwsli_provider = dict()
        self.line = tokens[0]
        self.nwsli = nwsli_provider.get(tokens[1], NWSLI(tokens[1]))
        self.severity = tokens[2]
        self.cause = tokens[3]
        self.beginTS = contime(tokens[4])
        self.crestTS = contime(tokens[5])
        self.endTS = contime(tokens[6])
        self.record = tokens[7]

    def __str__(self):
        """string representation of this HVTEC"""
        return self.line



"""Off-loaded private stuff from `vtec.py`."""
# pylint: disable=too-many-arguments
from datetime import timedelta
import itertools

import pandas as pd
from pyiem.util import LOG

# When a VTEC product has an infinity time 000000T0000Z, we need some value
# for the database to make things logically work.  We arb pick 21 days, which
# seems to be enough time to ensure a WFO issues some followup statement.
DEFAULT_EXPIRE_DELTA = timedelta(hours=(21 * 24))


def which_year(txn, prod, segment, vtec):
    """ Figure out which table we should work against """
    if vtec.action in ["NEW"]:
        # Lets piggyback a check to see if this ETN has been reused?
        # Can this realiably be done?
        txn.execute(
            f"SELECT max(updated) from warnings_{prod.db_year} WHERE wfo = %s "
            "and eventid = %s and significance = %s and phenomena = %s",
            (vtec.office, vtec.etn, vtec.significance, vtec.phenomena),
        )
        row = txn.fetchone()
        if row["max"] is not None:
            if (prod.valid - row["max"]).total_seconds() > (21 * 86400):
                prod.warnings.append(
                    "Possible Duplicated ETN\n"
                    f"  max(updated) is {row['max']}, "
                    f"prod.valid is {prod.valid}\n year is {prod.db_year}\n"
                    f"  VTEC: {str(vtec)}\n  "
                    f"product_id: {prod.get_product_id()}"
                )
        return prod.db_year
    # Lets query the database to look for any matching entries within
    # the past 3, 10, 31 days, to find with the product_issue was,
    # which guides the table that the data is stored within
    for offset in [3, 10, 31]:
        txn.execute(
            "SELECT tableoid::regclass as tablename, hvtec_nwsli, "
            "min(product_issue at time zone 'UTC'), "
            "max(product_issue at time zone 'UTC') from warnings "
            "WHERE wfo = %s and eventid = %s and significance = %s and "
            "phenomena = %s and ((updated > %s and updated <= %s) "
            "or expire > %s) and status not in ('UPG', 'CAN') "
            "GROUP by tablename, hvtec_nwsli ORDER by tablename DESC ",
            (
                vtec.office,
                vtec.etn,
                vtec.significance,
                vtec.phenomena,
                prod.valid - timedelta(days=offset),
                prod.valid,
                prod.valid,
            ),
        )
        rows = txn.fetchall()
        if not rows:
            continue
        if len(rows) > 1:
            # We likely have a flood warning and can use the HVTEC NWSLI
            # to resolve ambiguity
            hvtec_nwsli = segment.get_hvtec_nwsli()
            if hvtec_nwsli:
                for row in rows:
                    if hvtec_nwsli == row["hvtec_nwsli"]:
                        return int(row["tablename"].replace("warnings_", ""))

            prod.warnings.append(
                (
                    "VTEC %s product: %s returned %s rows when "
                    "searching for current table"
                )
                % (str(vtec), prod.get_product_id(), txn.rowcount)
            )
        row = rows[0]
        if row["min"] is not None:
            year = row["min"].year
            if row["max"].year != year:
                LOG.info(
                    "VTEC Product appears to cross 1 Jan UTC "
                    "minyear: %s maxyear: %s VTEC: %s productid: %s",
                    year,
                    row["max"].year,
                    str(vtec),
                    prod.get_product_id(),
                )
            return int(row["tablename"].replace("warnings_", ""))

    # Give up
    if not prod.is_correction():
        table = f"warnings_{prod.db_year}"
        prod.warnings.append(
            "Failed to find year of product issuance:\n"
            f"  VTEC:{str(vtec)}\n  PRODUCT: {prod.get_product_id()}\n"
            f"  defaulting to use year: {prod.db_year}\n"
            f"  {list_rows(txn, table, vtec)}"
        )
    return prod.db_year


def _associate_vtec_year(prod, txn):
    """Figure out to which year each VTEC in the product belongs.

    Modifies the prod.segment.vtec objects."""
    for seg, _ugcs, vtec in prod.suv_iter():
        if vtec.year is None:
            vtec.year = which_year(txn, prod, seg, vtec)


def _load_database_status(txn, prod):
    """Build a pandas dataframe for what the database knows."""
    rows = []
    done = []
    for _seg, _ugcs, vtec in prod.suv_iter():
        if vtec.status == "NEW" or vtec.year is None:
            continue
        key = f"{vtec.office}.{vtec.phenomena}.{vtec.significance}.{vtec.etn}"
        if key in done:
            continue
        done.append(key)
        txn.execute(
            "SELECT ugc, status, updated at time zone 'UTC' as utc_updated, "
            "expire at time zone 'UTC' as utc_expire "
            f"from warnings_{vtec.year} WHERE wfo = %s and "
            "phenomena = %s and significance = %s and eventid = %s and "
            "status not in ('CAN', 'UPG', 'EXP') and expire >= %s",
            (
                vtec.office,
                vtec.phenomena,
                vtec.significance,
                vtec.etn,
                prod.valid,
            ),
        )
        for row in txn.fetchall():
            entry = {
                "ugc": row[0],
                "status": row[1],
                "year": vtec.year,
                "phenomena": vtec.phenomena,
                "significance": vtec.significance,
                "etn": vtec.etn,
                "updated": row[2],
                "expire": row[3],
            }
            rows.append(entry)
    return pd.DataFrame(rows)


def check_dup_ps(segment):
    """Does this TextProductSegment have duplicated VTEC

    NWS AWIPS Developer asked that alerts be made when a VTEC segment has a
    phenomena and significance that are reused. In practice, this error is
    in the case of having the same phenomena.significance overlap in time. The
    combination of the same pheom.sig for events happening now and in the
    future is OK and common

    Returns:
      bool
    """
    combos = {}
    for thisvtec in segment.vtec:
        if thisvtec.begints is None or thisvtec.endts is None:
            # The logic here is too difficult for now, so we ignore
            continue
        key = thisvtec.s2()
        val = combos.setdefault(key, [])
        # we can't use vtec.endts in this situation
        endts = (
            segment.tp.valid
            if thisvtec.status in ["UPG", "CAN"]
            else thisvtec.endts
        )
        val.append([thisvtec.begints, endts])

    for key in combos:
        if len(combos[key]) == 1:
            continue
        for one, two in itertools.permutations(combos[key], 2):
            # We check for overlap
            if one[0] >= two[0] and one[0] < two[1]:
                return True
    return False


def do_sql_hvtec(txn, segment):
    """ Process the HVTEC in this product """
    nwsli = segment.hvtec[0].nwsli.id
    # No point in saving these events
    if nwsli == "00000":
        return
    if len(segment.bullets) < 4:
        return
    stage_text = ""
    flood_text = ""
    forecast_text = ""
    impact_text = ""
    for _, bullet in enumerate(segment.bullets):
        bsu = bullet.strip().upper()
        if bsu.find("FLOOD STAGE") == 0:
            flood_text = bullet
        if bsu.find("FORECAST") == 0:
            forecast_text = bullet
        if bsu.find("AT ") == 0 and stage_text == "":
            stage_text = bullet
        if bsu.startswith("IMPACT..."):
            impact_text = bullet.strip()[9:]

    txn.execute(
        "INSERT into riverpro(nwsli, stage_text, flood_text, forecast_text, "
        "impact_text, severity) VALUES (%s,%s,%s,%s,%s,%s)",
        (
            nwsli,
            stage_text,
            flood_text,
            forecast_text,
            impact_text,
            segment.hvtec[0].severity,
        ),
    )


def list_rows(txn, table, vtec):
    """Return a simple listing of what exists in the database"""
    txn.execute(
        (
            "SELECT ugc, issue at time zone 'UTC' as ui, status, "
            f"updated at time zone 'UTC' as uu from {table} "
            "WHERE wfo = %s and phenomena = %s and significance = %s and "
            "eventid = %s ORDER by ugc"
        ),
        (vtec.office, vtec.phenomena, vtec.significance, vtec.etn),
    )
    res = (
        f"Entries for VTEC within {table}\n"
        "  UGC    STA ISSUED              UPDATED\n"
    )
    for row in txn.fetchall():
        res += f"  {row['ugc']} {row['status']} {row['ui']} {row['uu']}\n"
    return res


def _debug_warning(prod, txn, warning_table, vtec, segment, ets):
    """ Get a more useful warning message for this failure """
    cnt = txn.rowcount
    txn.execute(
        "SELECT ugc, issue at time zone 'UTC' as utc_issue, "
        "expire at time zone 'UTC' as utc_expire, "
        "updated at time zone 'UTC' as utc_updated, "
        f"status from {warning_table} WHERE wfo = %s and eventid = %s and "
        "ugc in %s and significance = %s and phenomena = %s "
        "ORDER by ugc ASC, issue ASC",
        (
            vtec.office,
            vtec.etn,
            segment.get_ugcs_tuple(),
            vtec.significance,
            vtec.phenomena,
        ),
    )
    debugmsg = "UGC    STA ISSUE            EXPIRE           UPDATED\n"

    def myfmt(val):
        """ Be more careful """
        default = "%-16s" % ("((NULL))",)
        return default if val is None else val.strftime("%Y-%m-%d %H:%M")

    for row in txn.fetchall():
        debugmsg += (
            f"{row['ugc']} {row['status']} {myfmt(row['utc_issue'])} "
            f"{myfmt(row['utc_expire'])} {myfmt(row['utc_updated'])}\n"
        )
    return (
        f"Warning: {vtec.s3()} do_sql_vtec {warning_table} {vtec.action} "
        f"updated {cnt} row, should {len(segment.ugcs)} rows\n"
        f"UGCS: {segment.ugcs}\n"
        f"valid: {prod.valid} expire: {ets}\n{debugmsg}"
    )


def _resent_match(prod, txn, warning_table, vtec):
    """Check if this is a resent match."""
    txn.execute(
        f"SELECT max(updated) as maxtime from {warning_table} "
        "WHERE eventid = %s and significance = %s and wfo = %s and "
        "phenomena = %s",
        (vtec.etn, vtec.significance, vtec.office, vtec.phenomena),
    )
    maxtime = txn.fetchone()["maxtime"]
    if maxtime is not None and maxtime == prod.valid:
        LOG.info("RESENT Match, skipping SQL for %s!", prod.get_product_id())
        return True
    return False


def _do_sql_vtec_new(prod, txn, warning_table, segment, vtec):
    """ Do the NEW style actions."""
    bts = prod.valid if vtec.begints is None else vtec.begints
    # If this product has no expiration time, but db needs a value
    ets = vtec.endts
    if vtec.endts is None:
        ets = bts + DEFAULT_EXPIRE_DELTA

    fcster = prod.get_signature()
    if fcster is not None:
        fcster = fcster[:24]

    # For each UGC code in this segment, we create a database entry
    for ugc in segment.ugcs:
        # Check to see if we have entries already for this UGC
        # Some previous entries may not be in a terminated state, so
        # also check the expiration time
        txn.execute(
            f"SELECT issue, expire, updated from {warning_table} "
            "WHERE ugc = %s and eventid = %s and significance = %s "
            "and wfo = %s and phenomena = %s and "
            "status not in ('CAN', 'UPG') and expire > %s",
            (
                str(ugc),
                vtec.etn,
                vtec.significance,
                vtec.office,
                vtec.phenomena,
                prod.valid,
            ),
        )
        if txn.rowcount > 0:
            if prod.is_correction():
                # We'll delete old entries, gulp
                txn.execute(
                    f"DELETE from {warning_table} WHERE ugc = %s "
                    "and eventid = %s and significance = %s and "
                    "wfo = %s and phenomena = %s and "
                    "status in ('NEW', 'EXB', 'EXA') ",
                    (
                        str(ugc),
                        vtec.etn,
                        vtec.significance,
                        vtec.office,
                        vtec.phenomena,
                    ),
                )
                if txn.rowcount != 1:
                    prod.warnings.append(
                        f"{vtec.s3()} {str(ugc)} duplicated via "
                        f"product correction, deleted {txn.rowcount} "
                        "old rows instead of 1"
                    )

            else:
                prod.warnings.append(
                    "Duplicate(s) WWA found, "
                    f"rowcount: {txn.rowcount} for UGC: {ugc}"
                )

        txn.execute(
            f"INSERT into {warning_table} (issue, expire, updated, "
            "wfo, eventid, status, fcster, report, ugc, phenomena, "
            "significance, gid, init_expire, product_issue, "
            "hvtec_nwsli, hvtec_severity, hvtec_cause, hvtec_record, "
            "is_emergency, is_pds) "
            "VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, "
            "get_gid(%s, %s), %s, %s, %s, %s, %s, %s, %s, %s) RETURNING gid",
            (
                bts,
                ets,
                prod.valid,
                vtec.office,
                vtec.etn,
                vtec.action,
                fcster,
                prod.unixtext,
                str(ugc),
                vtec.phenomena,
                vtec.significance,
                str(ugc),
                prod.valid,
                ets,
                prod.valid,
                segment.get_hvtec_nwsli(),
                segment.get_hvtec_severity(),
                segment.get_hvtec_cause(),
                segment.get_hvtec_record(),
                segment.is_emergency,
                segment.is_pds,
            ),
        )
        # For unit tests, these mostly get filtered out
        if txn.fetchone().get("gid") is None:
            prod.warnings.append(f"get_gid({str(ugc)},{prod.valid}) was null")


def _do_sql_vtec_cor(prod, txn, warning_table, segment, vtec):
    """A Product Correction."""
    # A previous issued product is being corrected
    txn.execute(
        f"UPDATE {warning_table} SET "
        "expire = coalesce(%s, expire), status = %s, "
        "svs = (CASE WHEN (svs IS NULL) THEN '__' ELSE svs END) "
        "|| %s || '__', issue = coalesce(%s, issue), "
        "init_expire = coalesce(%s, init_expire) WHERE wfo = %s and "
        f"eventid = %s and ugc in %s and significance = %s "
        "and phenomena = %s and (expire + '1 hour'::interval) >= %s ",
        (
            vtec.endts,
            vtec.action,
            prod.unixtext,
            vtec.begints,
            vtec.endts,
            vtec.office,
            vtec.etn,
            segment.get_ugcs_tuple(),
            vtec.significance,
            vtec.phenomena,
            prod.valid,
        ),
    )
    if txn.rowcount != len(segment.ugcs):
        prod.warnings.append(
            _debug_warning(prod, txn, warning_table, vtec, segment, vtec.endts)
        )


def _do_sql_vtec_can(prod, txn, warning_table, segment, vtec):
    """A Product Correction."""
    ets = vtec.endts
    # These are terminate actions, so we act accordingly
    if vtec.action in ["CAN", "UPG"]:
        ets = prod.valid
    # If we are extending into infinity, but need a value
    if vtec.action == "EXT" and vtec.endts is None:
        ets = prod.valid + DEFAULT_EXPIRE_DELTA

    # An EXT action could change the issuance time, gasp
    issuesql = ""
    if vtec.action == "EXT" and vtec.begints is not None:
        issuesql = " issue = '%s', " % (vtec.begints,)
    txn.execute(
        f"UPDATE {warning_table} SET {issuesql} expire = %s, "
        "status = %s, updated = %s, "
        "svs = (CASE WHEN (svs IS NULL) THEN '__' ELSE svs END) "
        "|| %s || '__' WHERE wfo = %s and eventid = %s and ugc in "
        f"%s and significance = %s and phenomena = %s "
        "and status not in ('CAN', 'UPG') and "
        "(expire + '1 hour'::interval) >= %s",
        (
            ets,
            vtec.action,
            prod.valid,
            prod.unixtext,
            vtec.office,
            vtec.etn,
            segment.get_ugcs_tuple(),
            vtec.significance,
            vtec.phenomena,
            prod.valid,
        ),
    )
    if txn.rowcount != len(segment.ugcs):
        if not prod.is_correction():
            prod.warnings.append(
                _debug_warning(prod, txn, warning_table, vtec, segment, ets)
            )


def _do_sql_vtec_con(prod, txn, warning_table, segment, vtec):
    """Continue."""
    # These are no-ops, just updates
    ets = vtec.endts
    if vtec.endts is None:
        ets = prod.valid + DEFAULT_EXPIRE_DELTA

    # Offices have 1 hour to expire something :), actually 30 minutes
    txn.execute(
        f"UPDATE {warning_table} SET status = %s, updated = %s, "
        "svs = (CASE WHEN (svs IS NULL) THEN '__' ELSE svs END) "
        "|| %s || '__' , expire = %s, "
        "is_emergency = (case when %s then true else is_emergency end), "
        "is_pds = (case when %s then true else is_pds end) "
        f"WHERE wfo = %s and eventid = %s and ugc in %s "
        "and significance = %s and phenomena = %s and "
        "status not in ('CAN', 'UPG') and "
        "(expire + '1 hour'::interval) >= %s",
        (
            vtec.action,
            prod.valid,
            prod.unixtext,
            ets,
            segment.is_emergency,
            segment.is_pds,
            vtec.office,
            vtec.etn,
            segment.get_ugcs_tuple(),
            vtec.significance,
            vtec.phenomena,
            prod.valid,
        ),
    )
    if txn.rowcount != len(segment.ugcs):
        prod.warnings.append(
            _debug_warning(prod, txn, warning_table, vtec, segment, ets)
        )



"""Encapsulates a text product holding METARs."""
import re
from datetime import timezone, timedelta

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from metar.Metar import Metar
from metar.Metar import ParserError as MetarParserError
from pyiem.nws.product import TextProduct
from pyiem.observation import Observation
from pyiem.reference import TRACE_VALUE, TWEET_CHARS
from pyiem import datatypes
from pyiem.util import drct2text, LOG

NIL_RE = re.compile(r"[\s\n]NIL")
ERROR_RE = re.compile("Unparsed groups in body '(?P<msg>.*)' while processing")
TORNADO_RE = re.compile(r" \+FC |TORNADO")
FUNNEL_RE = re.compile(r" FC |FUNNEL")
# Match what looks like SA formatted messages
SA_RE = re.compile(r"^[A-Z]{3}\sSA")
# Sites we should route to Jabber
JABBER_SITES = {}
# Keep track of Wind alerts to prevent dups
WIND_ALERTS = {}
# Wind speed threshold in kts for alerting
WIND_ALERT_THRESHOLD_KTS = 50.0


def wind_logic(iem, this):
    """Hairy logic for now we handle winds."""
    # Explicit storages
    if this.wind_speed:
        iem.data["sknt"] = this.wind_speed.value("KT")
    if this.wind_gust:
        iem.data["gust"] = this.wind_gust.value("KT")
    if this.wind_dir:
        if this.wind_dir.value() == "VRB":
            iem.data["drct"] = 0
        else:
            iem.data["drct"] = float(this.wind_dir.value())
    if this.wind_speed_peak:
        iem.data["peak_wind_gust"] = this.wind_speed_peak.value("KT")
    if this.wind_dir_peak:
        iem.data["peak_wind_drct"] = this.wind_dir_peak.value()
    if this.peak_wind_time:
        iem.data["peak_wind_time"] = this.peak_wind_time.replace(
            tzinfo=timezone.utc
        )

    # Figure out if we have a new max_drct
    old_max_wind = max(
        [iem.data.get("max_sknt", 0) or 0, iem.data.get("max_gust", 0) or 0]
    )
    new_max_wind = max(
        [iem.data.get("sknt", 0) or 0, iem.data.get("gust", 0) or 0]
    )
    # if our sknt or gust is a new max, use drct
    if new_max_wind > old_max_wind:
        iem.data["max_drct"] = iem.data.get("drct", 0)
    # if our PK WND is greater than all yall, use PK WND
    # TODO: PK WND potentially could be from last hour / thus yesterday?
    if (
        this.wind_speed_peak
        and this.wind_dir_peak
        and this.wind_speed_peak.value("KT") > old_max_wind
        and this.wind_speed_peak.value("KT") > new_max_wind
    ):
        iem.data["max_drct"] = this.wind_dir_peak.value()
        iem.data["max_gust_ts"] = this.peak_wind_time.replace(
            tzinfo=timezone.utc
        )
        iem.data["max_gust"] = this.wind_speed_peak.value("KT")


def trace(pobj):
    """Convert this precip object to a numeric value"""
    if pobj is None:
        return None
    val = pobj.value("IN")
    if val == 0:
        # IEM denotation of trace
        return TRACE_VALUE
    return val


def to_metar(textprod, text):
    """Create a METAR object, if possible"""
    # Do some cleaning and whitespace trimming
    text = sanitize(text)
    if len(text) < 10:
        return
    attempt = 1
    mtr = None
    original_text = text
    valid = textprod.valid
    while attempt < 6 and mtr is None:
        try:
            mtr = METARReport(text, month=valid.month, year=valid.year)
        except MetarParserError as inst:
            tokens = ERROR_RE.findall(str(inst))
            if tokens:
                if tokens[0] == text or text.startswith(tokens[0]):
                    if not SA_RE.match(text):
                        LOG.info(
                            "%s Aborting due to non-replace %s",
                            textprod.get_product_id(),
                            str(inst),
                        )
                    return
                # So tokens contains a series of groups that needs updated
                newtext = text
                for token in tokens[0].split():
                    newtext = newtext.replace(" %s" % (token,), "")
                if newtext != text:
                    text = newtext
                else:
                    LOG.info("unparsed groups regex fail: %s", inst)
            if str(inst).find("day is out of range for month") > -1:
                if valid.day < 10:
                    valid = valid.replace(day=1) - timedelta(days=1)
        attempt += 1

    if mtr is not None:
        # Attempt to figure out more things
        if mtr.station_id is None:
            LOG.info("Aborting due to station_id being None |%s|", text)
            return None
        if mtr.time is None:
            LOG.info("Aborting due to time being None |%s|", text)
            return None
        # don't allow data more than an hour into the future
        ceiling = (textprod.utcnow + timedelta(hours=1)).replace(tzinfo=None)
        if mtr.time > ceiling:
            # careful, we may have obs from the previous month
            if ceiling.day < 5 and mtr.time.day > 15:
                prevmonth = ceiling - timedelta(days=10)
                mtr.time = mtr.time.replace(
                    year=prevmonth.year, month=prevmonth.month
                )
            else:
                LOG.info(
                    "Aborting due to time in the future "
                    "ceiling: %s mtr.time: %s",
                    ceiling,
                    mtr.time,
                )
                return None
        mtr.code = original_text
        mtr.iemid = (
            mtr.station_id[-3:] if mtr.station_id[0] == "K" else mtr.station_id
        )
        mtr.network = textprod.nwsli_provider.get(mtr.iemid, dict()).get(
            "network"
        )
        mtr.tzname = textprod.nwsli_provider.get(mtr.iemid, dict()).get(
            "tzname"
        )
    return mtr


def sanitize(text):
    """Clean our text string with METAR data"""
    text = re.sub("\015", " ", text)
    # Remove any multiple whitespace, bad chars
    text = (
        text.encode("utf-8", "ignore")
        .replace(b"\xa0", b" ")
        .replace(b"\001", b"")
        .replace(b"\003", b"")
        .decode("utf-8", errors="ignore")
    )
    text = " ".join(text.strip().split())
    # Look to see that our METAR starts with A-Z
    if re.match("^[0-9]", text):
        tokens = text.split()
        text = " ".join(tokens[1:])
    return text


def _is_same_day(valid, tzname, hours=6):
    """Can we trust a six hour total?"""
    try:
        tzinfo = ZoneInfo(tzname)
    except Exception:
        return False
    lts = valid.astimezone(tzinfo)
    # TODO we should likely somehow compute this in standard time, shrug
    return lts.day == (lts - timedelta(hours=hours)).day


class METARReport(Metar):
    """Provide some additional functionality over baseline METAR"""

    def __init__(self, text, **kwargs):
        """Wrapper"""
        Metar.__init__(self, text, **kwargs)
        self.iemid = None
        self.network = None
        self.tzname = None

    def wind_message(self):
        """Convert this into a Jabber style message"""
        drct = 0
        sknt = 0
        time = self.time.replace(tzinfo=timezone.utc)
        if self.wind_gust:
            sknt = self.wind_gust.value("KT")
            if self.wind_dir:
                drct = self.wind_dir.value()
        if self.wind_speed_peak:
            v1 = self.wind_speed_peak.value("KT")
            d1 = self.wind_dir_peak.value()
            t1 = self.peak_wind_time.replace(tzinfo=timezone.utc)
            if v1 > sknt:
                sknt = v1
                drct = d1
                time = t1
        key = "%s;%s;%s" % (self.station_id, sknt, time)
        if key not in WIND_ALERTS:
            WIND_ALERTS[key] = 1
            speed = datatypes.speed(sknt, "KT")
            return ("gust of %.0f knots (%.1f mph) from %s @ %s") % (
                speed.value("KT"),
                speed.value("MPH"),
                drct2text(drct),
                time.strftime("%H%MZ"),
            )

    def over_wind_threshold(self):
        """Is this METAR over the wind threshold for alerting"""
        if (
            self.wind_gust
            and self.wind_gust.value("KT") >= WIND_ALERT_THRESHOLD_KTS
        ):
            return True
        if (
            self.wind_speed_peak
            and self.wind_speed_peak.value("KT") >= WIND_ALERT_THRESHOLD_KTS
        ):
            return True
        return False

    def to_iemaccess(self, txn, force_current_log=False, skip_current=False):
        """Persist parsed data to IEMAccess Database.

        Args:
          txn (psycopg2.cursor): database cursor / transaction
          force_current_log (boolean): should this ob always go to current_log
          skip_current (boolean): should this ob always skip current table
        """
        gts = self.time.replace(tzinfo=timezone.utc)
        iem = Observation(self.iemid, self.network, gts)
        # Load the observation from the database, if the same time exists!
        iem.load(txn)

        # Need to figure out if we have a duplicate ob, if so, check
        # the length of the raw data, if greater, take the temps
        if iem.data["raw"] is None or len(iem.data["raw"]) < len(self.code):
            if self.temp:
                val = self.temp.value("F")
                # Place reasonable bounds on the temperature before saving it!
                if val > -90 and val < 150:
                    iem.data["tmpf"] = round(val, 1)
            if self.dewpt:
                val = self.dewpt.value("F")
                # Place reasonable bounds on the temperature before saving it!
                if val > -150 and val < 100:
                    iem.data["dwpf"] = round(val, 1)
            # Database only allows len 254
            iem.data["raw"] = self.code[:254]
        # Always take a COR
        if self.code.find(" COR ") > -1:
            iem.data["raw"] = self.code[:254]

        wind_logic(iem, self)

        if self.max_temp_6hr:
            iem.data["max_tmpf_6hr"] = round(self.max_temp_6hr.value("F"), 1)
            if self.tzname and _is_same_day(iem.data["valid"], self.tzname):
                iem.data["max_tmpf_cond"] = iem.data["max_tmpf_6hr"]
        if self.min_temp_6hr:
            iem.data["min_tmpf_6hr"] = round(self.min_temp_6hr.value("F"), 1)
            if self.tzname and _is_same_day(iem.data["valid"], self.tzname):
                iem.data["min_tmpf_cond"] = iem.data["min_tmpf_6hr"]
        if self.max_temp_24hr:
            iem.data["max_tmpf_24hr"] = round(self.max_temp_24hr.value("F"), 1)
        if self.min_temp_24hr:
            iem.data["min_tmpf_24hr"] = round(self.min_temp_24hr.value("F"), 1)
        if self.precip_3hr:
            iem.data["p03i"] = trace(self.precip_3hr)
        if self.precip_6hr:
            iem.data["p06i"] = trace(self.precip_6hr)
        if self.precip_24hr:
            iem.data["p24i"] = trace(self.precip_24hr)
        # We assume the value is zero, sad!
        iem.data["phour"] = 0
        if self.precip_1hr:
            iem.data["phour"] = trace(self.precip_1hr)

        if self.snowdepth:
            iem.data["snowd"] = self.snowdepth.value("IN")
        if self.vis:
            iem.data["vsby"] = self.vis.value("SM")
        if self.press:
            iem.data["alti"] = self.press.value("IN")
        if self.press_sea_level:
            iem.data["mslp"] = self.press_sea_level.value("MB")
        if self.press_sea_level and self.press:
            alti = self.press.value("MB")
            mslp = self.press_sea_level.value("MB")
            if abs(alti - mslp) > 25:
                LOG.info(
                    "PRESSURE ERROR %s %s ALTI: %s MSLP: %s",
                    iem.data["station"],
                    iem.data["valid"],
                    alti,
                    mslp,
                )
                if alti > mslp:
                    iem.data["mslp"] += 100.0
                else:
                    iem.data["mslp"] -= 100.0
        # Do something with sky coverage
        for i in range(len(self.sky)):
            (cov, hgh, _) = self.sky[i]
            iem.data["skyc%s" % (i + 1)] = cov
            if hgh is not None:
                iem.data["skyl%s" % (i + 1)] = hgh.value("FT")

        # Presentwx
        if self.weather:
            pwx = []
            for wx in self.weather:
                val = "".join([a for a in wx if a is not None])
                if val == "" or val == len(val) * "/":
                    continue
                pwx.append(val)
            iem.data["wxcodes"] = pwx

        # Ice Accretion
        for hr in [1, 3, 6]:
            key = "ice_accretion_%shr" % (hr,)
            iem.data[key] = trace(getattr(self, key))
        return iem, iem.save(txn, force_current_log, skip_current)


class METARCollective(TextProduct):
    """
    A TextProduct containing METAR information
    """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """Constructor

        Args:
          text (string): the raw string to process"""
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)
        self.metars = []
        self.split_and_parse()

    def get_jabbers(self, uri, _uri2=None):
        """Make this into jabber messages"""
        jmsgs = []
        for mtr in self.metars:
            msg = None
            for weatheri in mtr.weather:
                for wx in weatheri:
                    if wx is not None and "GR" in wx:
                        msg = "Hail"
            if TORNADO_RE.findall(mtr.code):
                msg = "Tornado"
            elif FUNNEL_RE.findall(mtr.code):
                msg = "Funnel Cloud"
            # Search for Peak wind gust info....
            elif mtr.over_wind_threshold():
                _msg = mtr.wind_message()
                if _msg:
                    msg = _msg
            elif mtr.station_id in JABBER_SITES:
                # suck
                if JABBER_SITES[mtr.station_id] != mtr.time:
                    JABBER_SITES[mtr.station_id] = mtr.time
                    channels = ["METAR.%s" % (mtr.station_id,)]
                    if mtr.type == "SPECI":
                        channels.append("SPECI.%s" % (mtr.station_id,))
                    mstr = "%s %s" % (mtr.type, mtr.code)
                    jmsgs.append(
                        [mstr, mstr, dict(channels=",".join(channels))]
                    )
            if msg:
                row = self.nwsli_provider.get(mtr.iemid, dict())
                wfo = row.get("wfo")
                if wfo is None or wfo == "":
                    LOG.info(
                        "Unknown WFO for id: %s, skipping alert", mtr.iemid
                    )
                    continue
                channels = ["METAR.%s" % (mtr.station_id,)]
                if mtr.type == "SPECI":
                    channels.append("SPECI.%s" % (mtr.station_id,))
                channels.append(wfo)
                st = row.get("state")
                nm = row.get("name")

                extra = ""
                if mtr.code.find("$") > 0:
                    extra = "(Caution: Maintenance Check Indicator)"
                url = ("%s%s") % (uri, mtr.network)
                jtxt = ("%s,%s (%s) ASOS %s reports %s\n%s %s") % (
                    nm,
                    st,
                    mtr.iemid,
                    extra,
                    msg,
                    mtr.code,
                    url,
                )
                jhtml = (
                    f'<p><a href="{url}">{nm},{st}</a> ({mtr.iemid}) ASOS '
                    f"{extra} reports <strong>{msg}</strong>"
                    f"<br/>{mtr.code}</p>"
                )
                xtra = {
                    "channels": ",".join(channels),
                    "lat": str(row.get("lat")),
                    "long": str(row.get("lon")),
                }
                xtra["twitter"] = (
                    ("%s,%s (%s) ASOS reports %s -- %s")
                    % (nm, st, mtr.iemid, msg, mtr.code)
                )[:TWEET_CHARS]
                jmsgs.append([jtxt, jhtml, xtra])

        return jmsgs

    def split_and_parse(self):
        """Create METAR objects as we find products in the text"""
        # skip the top three lines
        lines = self.unixtext.split("\n")
        if lines[0] == "\001":
            content = "\n".join(lines[3:])
        elif len(lines[0]) < 5:
            content = "\n".join(lines[2:])
        else:
            self.warnings.append(
                ("WMO header split_and_parse fail: %s") % (self.unixtext,)
            )
            content = "\n".join(lines)
        # Tokenize on the '=', which splits a product with METARs
        tokens = content.split("=")
        for token in tokens:
            # Dump METARs that have NIL in them
            prefix = "METAR" if self.afos != "SPECI" else "SPECI"
            if NIL_RE.search(token):
                continue
            elif token.find("METAR") > -1:
                token = token[(token.find("METAR") + 5) :]
            # unsure why this LWIS existed
            # elif token.find("LWIS ") > -1:
            #    token = token[token.find("LWIS ")+5:]
            elif token.find("SPECI") > -1:
                token = token[(token.find("SPECI") + 5) :]
                prefix = "SPECI"
            elif len(token.strip()) < 5:
                continue
            res = to_metar(self, token)
            if res:
                res.type = prefix
                self.metars.append(res)


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Helper function """
    return METARCollective(text, utcnow, ugc_provider, nwsli_provider)



"""Generate the _ccw files used in pyIEM."""

from pyiem.plot import MapPlot
from pyiem.util import get_dbconn
from shapely.wkb import loads
from shapely.geometry import Polygon, MultiPolygon
from descartes.patch import PolygonPatch
import numpy as np


def main():
    """Go Main Go."""
    DBCONN = get_dbconn("postgis")
    cursor = DBCONN.cursor()

    cursor.execute(
        """
        SELECT
            ST_Transform(
                ST_Simplify(
                    ST_Union(ST_Transform(the_geom,2163)),
                    500.),
                4326) from states
        WHERE state_abbr in ('IA', 'IL', 'IN')
    ---    where state_abbr in ('MI','WI', 'IL', 'IN', 'OH', 'KY', 'MO', 'KS',
    ---    'NE', 'SD', 'ND', 'MN', 'IA')
    """
    )

    m = MapPlot("iowa")
    lons = np.linspace(-110, -70, 50)
    lats = np.linspace(28, 52, 50)
    vals = np.linspace(0, 50, 50)
    m.contourf(lons, lats, vals, vals)

    # """
    for row in cursor:
        multipoly = MultiPolygon([loads(row[0], hex=True)])
        for geo in multipoly.geoms:
            if geo.area < 1:
                continue
            (lons, lats) = geo.exterior.xy
            print("Masking with geo... %s %s" % (geo.area, len(lons)))

            ar = list(zip(lons, lats))
            ar.reverse()
            np.save("iailin_ccw.npy", ar)
            # mask_outside_polygon(ar, ax=m.ax)
            poly = PolygonPatch(Polygon(ar), fc="r", zorder=10000)
            m.ax.add_patch(poly)
    # """
    m.postprocess(filename="test.png")


if __name__ == "__main__":
    main()



"""SPS Parsing"""
# pylint: disable=redefined-outer-name

import pytest
from psycopg2.extras import RealDictCursor
from pyiem.util import get_dbconn, get_test_file
from pyiem.nws.ugc import UGC
from pyiem.nws.products import parser as spsparser


@pytest.fixture
def dbcursor():
    """Get a cursor we can use"""
    pgconn = get_dbconn("postgis")
    return pgconn.cursor(cursor_factory=RealDictCursor)


def test_sps(dbcursor):
    """Can we parse a SPS properly, yes we can!"""
    ugc_provider = {"ALZ039": UGC("AL", "Z", "039", name="Marengo")}
    prod = spsparser(get_test_file("SPSBMX.txt"), ugc_provider=ugc_provider)
    jmsgs = prod.get_jabbers("http://localhost")
    assert len(prod.segments) == 2
    assert len(jmsgs) == 1
    expected = (
        "<p>BMX issues <a href='http://localhost?pid=201805292152-"
        "KBMX-WWUS84-SPSBMX'>SIGNIFICANT WEATHER ADVISORY FOR "
        "SOUTHWESTERN MARENGO COUNTY UNTIL 515 PM CDT</a></p>"
    )
    assert jmsgs[0][1] == expected
    assert "SPSBMX" in jmsgs[0][2]["channels"]
    assert "SPS..." in jmsgs[0][2]["channels"]
    assert "SPSBMX.ALZ039" in jmsgs[0][2]["channels"]
    assert "ALZ039" in jmsgs[0][2]["channels"]

    prod.sql(dbcursor)
    dbcursor.execute(
        """
        SELECT count(*) from text_products where product_id = %s
    """,
        (prod.get_product_id(),),
    )
    assert dbcursor.fetchall()[0]["count"] == 1



"""HML"""
# pylint: disable=redefined-outer-name
import datetime

import psycopg2.extras
import pytest
from pyiem.nws.products.hml import parser as hmlparser
from pyiem.util import get_dbconn, get_test_file, utc


@pytest.fixture
def dbcursor():
    """Get database conn."""
    dbconn = get_dbconn("hml")
    # Note the usage of RealDictCursor here, as this is what
    # pyiem.twistedpg uses
    yield dbconn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    dbconn.close()


def test_200926_nokey(dbcursor):
    """Test that we add a new observation key, when necessary."""
    prod = hmlparser(get_test_file("HML/HMLMOB.txt"))
    prod.sql(dbcursor)
    dbcursor.execute(
        "SELECT key from hml_observed_data WHERE station = 'EFRA1' "
        "and valid >= '2020-09-26 13:36+00' and valid < '2020-09-26 13:43+00' "
        "and key is null"
    )
    assert dbcursor.rowcount == 0


def test_190313_missingstage(dbcursor):
    """Figure out why this HML is missing stage info."""
    prod = hmlparser(get_test_file("HML/HMLDMX.txt"))
    assert not prod.warnings
    prod.sql(dbcursor)
    dbcursor.execute(
        """
        SELECT * from hml_observed_data WHERE station = 'JANI4'
        and valid > '2019-03-13' and valid < '2019-03-14'
    """
    )
    assert dbcursor.rowcount == 8


def test_160826_hmlarx(dbcursor):
    """Lets dance"""
    utcnow = utc(2016, 8, 26, 8)
    prod = hmlparser(get_test_file("HML/HMLARX.txt"), utcnow=utcnow)
    prod.sql(dbcursor)
    assert not prod.warnings
    assert prod.data[0].stationname == "CEDAR RIVER 2 S St. Ansgar"


def test_161010_timing():
    """test how fast we can parse the file, over and over again"""
    sts = datetime.datetime.now()
    for _ in range(100):
        hmlparser(get_test_file("HML/HMLARX.txt"))
    ets = datetime.datetime.now()
    rate = (ets - sts).total_seconds() / 100.0
    print("sec per parse %.4f" % (rate,))
    assert rate < 1.0



"""Parsing of Flash Flood Guidances

NWS Discontinued 30 Sep 2018
https://www.weather.gov/media/notification/pdfs/pns18-13disc_county_ffg.pdf
"""
import re
from datetime import timezone, datetime

import pandas as pd
from pyiem.nws.product import TextProduct

SHEFRE = re.compile(
    (
        r"\.B (?P<src>[A-Z]{3}) (?P<date>[0-9]{6,8}) Z "
        r"DH(?P<hh>[0-9]{2})/DC(?P<valid>[0-9]{10,12}) "
        r"/DUE/PPHCF/PPTCF/PPQCF"
    )
)
DATARE = re.compile(
    (
        r"^(?P<ugc>[A-Z0-9]{6})\s+(?P<hour01>[0-9\.]+)/\s+"
        r"(?P<hour03>[0-9\.]+)/\s+(?P<hour06>[0-9\.]+)\s*"
        r"/?\s*(?P<hour12>[0-9\.]+)?\s*"
        r"/?\s*(?P<hour24>[0-9\.]+)?\s*"
    ),
    re.M,
)


def safe(val):
    """Safe conversion to float"""
    if val is None:
        return None
    return float(val)


class FFGProduct(TextProduct):
    """Class representing a FFG Product"""

    def __init__(self, text, utcnow=None):
        """Constructor

        Args:
          text (str): text to parse
        """
        TextProduct.__init__(self, text, utcnow=utcnow)
        self.data = None
        self.issue = None
        self.do_parsing()

    def do_parsing(self):
        """Process this file and save data"""
        shef = SHEFRE.search(self.text)
        if shef is None:
            self.warnings.append("Failed to find SHEF variable!")
            return
        group = shef.groupdict()
        self.issue = datetime.strptime(group["date"][-6:], "%y%m%d").replace(
            tzinfo=timezone.utc
        )
        self.issue = self.issue.replace(hour=(int(group["hh"]) % 24))
        dc = datetime.strptime(group["valid"][-10:], "%y%m%d%H%M").replace(
            tzinfo=timezone.utc
        )
        # Emailed KTUA about this on 17 Apr 2017
        if (
            abs((self.issue - dc).total_seconds()) > (12 * 3600.0)
            and self.source != "KTUA"
        ):
            self.warnings.append(
                "Product has large delta between DC: "
                f"{dc.strftime('%Y-%m-%d %H:%MZ')} and "
                f"SHEF Date: {self.issue.strftime('%Y-%m-%d %H:%MZ')}"
            )
        rows = []
        pos1 = self.unixtext.find(".B ")
        pos2 = self.unixtext.find(".END")
        if pos1 == -1 or pos2 == -1:
            return
        for match in DATARE.finditer(self.unixtext[pos1:pos2]):
            group = match.groupdict()
            rows.append(
                dict(
                    ugc=group["ugc"],
                    hour01=safe(group["hour01"]),
                    hour03=safe(group["hour03"]),
                    hour06=safe(group["hour06"]),
                    hour12=safe(group["hour12"]),
                    hour24=safe(group["hour24"]),
                )
            )
        self.data = pd.DataFrame(rows)

    def sql(self, txn):
        """Do the necessary database work

        Args:
          (psycopg2.transaction): a database transaction
        """
        if self.data is None:
            self.warnings.append("sql() was called with no data parsed!")
            return
        table = "ffg_%s" % (self.issue.year,)
        for _, row in self.data.iterrows():
            txn.execute(
                f"INSERT into {table} (ugc, valid, hour01, hour03, hour06, "
                "hour12, hour24) VALUES (%s, %s, %s, %s, %s, %s, %s)",
                (
                    row["ugc"],
                    self.issue,
                    row["hour01"],
                    row["hour03"],
                    row["hour06"],
                    row["hour12"],
                    row["hour24"],
                ),
            )


def parser(text, utcnow=None):
    """parser of raw SPC SAW Text

    Args:
      text (str): the raw text to parse
      utcnow (datetime): the current datetime with timezone set!

    Returns:
      SAWProduct instance
    """
    return FFGProduct(text, utcnow=utcnow)



"""Supports parsing of the NWNFormat

Which is a format used by the Texas Weather Sensors KCCI-TV Operates
"""
from datetime import timezone, datetime
import traceback
import re
import math

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

import pyiem.reference as reference
import pyiem.util as util


def uv(sped, drct2):
    """Convert to u,v components."""
    dirr = drct2 * math.pi / 180.00
    s = math.sin(dirr)
    c = math.cos(dirr)
    u = round(-sped * s, 2)
    v = round(-sped * c, 2)
    return u, v


def dwpf(tmpf, relh):
    """
    Compute the dewpoint in F given a temperature and relative humidity
    """
    if tmpf is None or relh is None:
        return None

    tmpk = 273.15 + (5.00 / 9.00 * (tmpf - 32.00))
    dwpk = tmpk / (1 + 0.000425 * tmpk * -(math.log10(relh / 100.0)))
    return int(float((dwpk - 273.15) * 9.00 / 5.00 + 32))


def mydir(u, v):
    if v == 0:
        v = 0.000000001
    dd = math.atan(u / v)
    ddir = (dd * 180.00) / math.pi

    if u > 0 and v > 0:
        ddir = 180 + ddir
    elif u > 0 and v < 0:
        ddir = 360 + ddir
    elif u < 0 and v > 0:
        ddir = 180 + ddir
    return int(round(math.fabs(ddir), 0))


def feelslike(tmpf, relh, sped):
    if tmpf > 50:
        return heatidx(tmpf, relh)
    else:
        return wchtidx(tmpf, sped)


def heatidx(tmpf, relh):
    if tmpf < 70:
        return tmpf
    if tmpf > 140:
        return -99
    if relh == 0:
        return -99

    PR_HEAT = 61 + (tmpf - 68) * 1.2 + relh * 0.094
    if PR_HEAT < 77:
        return PR_HEAT

    t2 = tmpf * tmpf
    t3 = tmpf * tmpf * tmpf
    r2 = relh * relh
    r3 = relh * relh * relh

    return (
        17.423
        + 0.185212 * tmpf
        + 5.37941 * relh
        - 0.100254 * tmpf * relh
        + 0.00941695 * t2
        + 0.00728898 * r2
        + 0.000345372 * t2 * relh
        - 0.000814971 * tmpf * r2
        + 0.0000102102 * t2 * r2
        - 0.000038646 * t3
        + 0.0000291583 * r3
        + 0.00000142721 * t3 * relh
        + 0.000000197483 * tmpf * r3
        - 0.0000000218429 * t3 * r2
        + 0.000000000843296 * t2 * r3
        - 0.0000000000481975 * t3 * r3
    )


def wchtidx(tmpf, sped):
    if sped < 3 or tmpf > 50:
        return tmpf
    wci = math.pow(sped, 0.16)

    return 35.74 + 0.6215 * tmpf - 35.75 * wci + 0.4275 * tmpf * wci


class nwnformat:
    def __init__(self, do_avg_winds=True):
        self.error = 0
        self.do_avg_winds = do_avg_winds

        self.sid = None
        self.ts = None
        self.avg_sknt = None
        self.avg_drct = None
        self.drct = None
        self.drctTxt = None
        self.avg_drctTxt = None
        self.sped = None
        self.sknt = None
        self.rad = 0
        self.insideTemp = 460
        self.tmpf = None
        self.humid = None
        self.pres = None
        self.presTend = None
        self.pDay = 0.00
        self.pMonth = 0.00
        self.pRate = 0.00
        self.dwpf = None
        self.feel = None

        self.nhumid = 0
        self.xhumid = 0

        self.npres = 0
        self.xpres = 0

        self.xtmpf = None
        self.ntmpf = None
        self.xsped = None
        self.xdrct = None
        self.xdrctTxt = None
        self.xsrad = None

        self.strMaxLine = None
        self.strMinLine = None

        self.aSknt = []
        self.aDrct = []

    def setTS(self, newval):
        try:
            self.ts = datetime.strptime(newval, "%m/%d/%y %H:%M:%S")
        except Exception:
            traceback.print_exc()
            self.error = 100
            return
        now = datetime.now()
        if (now - self.ts).total_seconds() > 7200:
            self.error = 101

    def avgWinds(self):
        if not self.aSknt:
            self.sped = None
            self.drct = None
            return

        self.avg_sknt = int(float(sum(self.aSknt)) / float(len(self.aSknt)))
        utot = 0
        vtot = 0
        for s, d in zip(self.aSknt, self.aDrct):
            u, v = uv(s, d)
            if s > self.xsped:
                self.xsped = s * 1.150
                self.xdrct = d
                self.xdrctTxt = util.drct2text(d)

            utot += u
            vtot += v
        uavg = utot / float(len(self.aSknt))
        vavg = vtot / float(len(self.aSknt))
        self.avg_drct = mydir(uavg, vavg)
        self.avg_drctTxt = util.drct2text(self.avg_drct)

        self.aSknt = []
        self.aDrct = []

    def parseLineRT(self, tokens):
        if self.ts is None:
            _t = datetime.utcnow()
            _t = _t.replace(second=0, microsecond=0, tzinfo=timezone.utc)
            self.ts = _t.astimezone(ZoneInfo("America/Chicago"))

        if len(tokens) != 14:
            return
        lineType = tokens[2]
        if lineType == "Max":
            self.parseMaxLineRT(tokens)
        elif lineType == "Min":
            self.parseMinLineRT(tokens)
        else:
            _t = datetime.utcnow()
            _t = _t.replace(second=0, microsecond=0, tzinfo=timezone.utc)
            self.ts = _t.astimezone(ZoneInfo("America/Chicago"))
            self.parseCurrentLineRT(tokens)

    def parseMaxLineRT(self, tokens):
        self.xdrct = reference.txt2drct[tokens[4]]
        self.xdrctTxt = tokens[4]
        if len(tokens[5]) >= 5:
            t = re.findall("([0-9]+)(MPH|KTS)", tokens[5])[0]
            if t[1] == "MPH":
                self.xsped = int(t[0])
            else:
                sknt = int(t[0])
                self.xsped = round(sknt * 1.1507, 0)

        if len(tokens[6]) == 4:
            self.xsrad = (
                int(re.findall("([0-9][0-9][0-9])[F,K]", tokens[6])[0]) * 10
            )

        if len(tokens[8]) == 4 or len(tokens[8]) == 3:
            if tokens[8][0] == "0":
                tokens[8] = tokens[8][1:]
            self.xtmpf = int(tokens[8][:-1])

    def parseMinLineRT(self, tokens):
        if len(tokens[8]) == 4 or len(tokens[8]) == 3:
            if tokens[8][0] == "0":
                tokens[8] = tokens[8][1:]
            self.ntmpf = int(tokens[8][:-1])

    def parseCurrentLineRT(self, tokens):
        # ['M', '057', '09:57', '09/04/03', 'ESE', '01MPH', '058K', '460F',
        #  '065F', '070%', '30.34R', '00.00"D', '00.00"M', '00.00"R']
        # Don't forget about this lovely one!
        # ['M', '057', '09:57', '09/04/03', 'ESE', '01MPH', '058K', '460F',
        #  '0-5F', '070%', '30.34R', '00.00"D', '00.00"M', '00.00"R']
        if len(tokens[8]) == 4 or len(tokens[8]) == 3:
            if tokens[8][0] == "0":
                tokens[8] = tokens[8][1:]
            self.tmpf = int(tokens[8][:-1])

        self.drct = reference.txt2drct[tokens[4]]
        self.drctTxt = tokens[4]
        if self.do_avg_winds:
            self.aDrct.append(int(self.drct))

        if len(tokens[5]) >= 5:
            t = re.findall("([0-9]+)(MPH|KTS)", tokens[5])[0]
            if t[1] == "MPH":
                self.sped = int(t[0])
                self.sknt = round(float(self.sped) * 0.868976, 0)
            else:
                self.sknt = int(t[0])
                self.sped = round(self.sknt / 0.868976, 0)
        if self.do_avg_winds:
            self.aSknt.append(self.sknt)

        if len(tokens[6]) == 4:
            self.rad = (
                int(re.findall("([0-9][0-9][0-9])[F,K]", tokens[6])[0]) * 10
            )

        if len(tokens[9]) == 4:
            self.humid = int(re.findall("([0-9][0-9][0-9])%", tokens[9])[0])

        if len(tokens[10]) == 6:
            self.pres = float(re.findall("(.*).", tokens[10])[0])

        if len(tokens[11]) == 7:
            self.pDay = float(re.findall('(.*)"D', tokens[11])[0])

        if len(tokens[12]) == 7:
            self.pMonth = float(re.findall('(.*)"M', tokens[12])[0])

        if (
            self.tmpf > -50
            and self.tmpf < 120
            and self.humid > 5
            and self.humid < 100.1
        ):
            self.dwpf = dwpf(self.tmpf, self.humid)
            self.feel = feelslike(self.tmpf, self.humid, self.sped)
        else:
            self.dwpf = None
            self.feel = None

    def currentLine(self):
        """Return NWN formatted string for current ob

        Returns:
          str: A NWN format string"""
        return (
            "%s %03.0f  %5s %8s %-3s %02.0fMPH %03.0fK %03.0fF %03.0fF "
            '%03.0f%s %05.2f%s %05.2f"D %05.2f"M %05.2f"R\015\012'
        ) % (
            "A",
            self.sid,
            self.ts.strftime("%H:%M"),
            self.ts.strftime("%m/%d/%y"),
            self.drctTxt,
            self.sped,
            self.rad,
            self.insideTemp,
            self.tmpf,
            self.humid,
            "%",
            self.pres,
            self.presTend,
            self.pDay,
            self.pMonth,
            self.pRate,
        )

    def maxLine(self):
        """Return NWN formatted string for max ob

        Returns:
          str: A NWN format string"""
        return (
            "%s %03i  %5s %8s %-3s %02iMPH %03iK %03iF %03iF "
            '%03i%s %05.2f%s %05.2f"D %05.2f"M %05.2f"R\015\012'
        ) % (
            "A",
            self.sid,
            "Max ",
            self.ts.strftime("%m/%d/%y"),
            "N",
            self.xsped,
            self.rad,
            self.insideTemp,
            self.xtmpf,
            self.xhumid,
            "%",
            self.xpres,
            self.presTend,
            0,
            0,
            0,
        )

    def minLine(self):
        """Return NWN formatted string for min ob

        Returns:
          str: A NWN format string"""
        return (
            "%s %03i  %5s %8s %-3s %02iMPH %03iK %03iF %03iF %03i%s "
            '%05.2f" %05.2f"D %05.2f"M %05.2f"R\015\012'
        ) % (
            "A",
            self.sid,
            "Min ",
            self.ts.strftime("%m/%d/%y"),
            self.drctTxt,
            0,
            0,
            self.insideTemp,
            self.ntmpf,
            self.nhumid,
            "%",
            self.npres,
            0,
            0,
            0,
        )

    def sanityCheck(self):
        """Bounds Check."""
        if self.xsped is None or self.xsped < 0:
            self.xsped = 0
            self.xdrct = -99

        if self.pres is None or self.pres < 0:
            self.pres = 0
        if self.pDay is None or self.pDay < 0:
            self.pDay = 0
        if self.humid is None or self.humid < 0 or self.humid > 100:
            self.humid = 0
        if self.tmpf is None or self.tmpf < -100 or self.tmpf > 150:
            self.tmpf = 460
        if self.ntmpf is None or self.ntmpf < -100 or self.ntmpf > 150:
            self.ntmpf = 460
        if self.xtmpf is None or self.xtmpf < -100 or self.xtmpf > 150:
            self.xtmpf = 460
        if self.sped is None or self.sped < 0 or self.sped > 300:
            self.sped = 0
        if self.avg_sknt is None or self.avg_sknt < 0 or self.avg_sknt > 300:
            self.avg_sknt = 0
        if self.avg_drct is None or self.avg_drct < 0 or self.avg_drct > 360:
            self.avg_drct = 0
        if self.xsrad is None or self.xsrad < 0 or self.xsrad > 10000:
            self.xsrad = 0



"""Tests for the pyiem.meteorology library"""
import warnings

import numpy as np
from metpy.units import units, masked_array
from pyiem import datatypes, meteorology

warnings.simplefilter("ignore", RuntimeWarning)


def test_vectorized():
    """See that heatindex and windchill can do lists"""
    temp = datatypes.temperature([0, 10], "F")
    sknt = datatypes.speed([30, 40], "MPH")
    val = meteorology.windchill(temp, sknt).value("F")
    assert abs(val[0] - -24.50) < 0.01

    t = datatypes.temperature([80.0, 90.0], "F")
    td = datatypes.temperature([70.0, 60.0], "F")
    hdx = meteorology.heatindex(t, td)
    assert abs(hdx.value("F")[0] - 83.93) < 0.01

    tmpf = np.array([80.0, 90.0]) * units("degF")
    dwpf = np.array([70.0, 60.0]) * units("degF")
    smps = np.array([10.0, 20.0]) * units("meter per second")
    feels = meteorology.mcalc_feelslike(tmpf, dwpf, smps)
    assert abs(feels.to(units("degF")).magnitude[0] - 83.15) < 0.01

    tmpf = masked_array([80.0, np.nan], units("degF"), mask=[False, True])
    feels = meteorology.mcalc_feelslike(tmpf, dwpf, smps)
    assert abs(feels.to(units("degF")).magnitude[0] - 83.15) < 0.01
    assert feels.mask[1]


def test_gdd_with_nans():
    """Can we properly deal with nan's and not emit warnings?"""
    highs = np.ma.array([70, 80, np.nan, 90], mask=[False, False, True, False])
    lows = highs - 10
    r = meteorology.gdd(
        datatypes.temperature(highs, "F"),
        datatypes.temperature(lows, "F"),
        50,
        86,
    )
    assert np.ma.is_masked(r[2])


def test_gdd():
    """Growing Degree Days"""
    r = meteorology.gdd(
        datatypes.temperature(86, "F"), datatypes.temperature(50, "F"), 50, 86
    )
    assert r == 18

    r = meteorology.gdd(
        datatypes.temperature(51, "F"), datatypes.temperature(49, "F"), 50, 86
    )
    assert abs(r - 0.5) < 0.1

    r = meteorology.gdd(
        datatypes.temperature(49, "F"), datatypes.temperature(40, "F"), 50, 86
    )
    assert r == 0

    r = meteorology.gdd(
        datatypes.temperature([86, 86], "F"),
        datatypes.temperature([50, 50], "F"),
        50,
        86,
    )
    assert r[0] == 18
    assert r[1] == 18


def test_mixingratio():
    """Test the mixing ratio calculation"""
    r = meteorology.mixing_ratio(datatypes.temperature(70, "F"))
    assert abs(r.value("KG/KG") - 0.016) < 0.001


def test_sw():
    """Test shortwave flux calculation"""
    r = meteorology.clearsky_shortwave_irradiance_year(42, 100)
    assert abs(r[0] - 7.20) < 0.01
    assert abs(r[90] - 22.45) < 0.01
    assert abs(r[182] - 32.74) < 0.01
    assert abs(r[270] - 19.07) < 0.01
    assert abs(r[364] - 7.16) < 0.01


def test_drct():
    """Conversion of u and v to direction"""
    r = meteorology.drct(
        datatypes.speed(np.array([10, 20]), "KT"),
        datatypes.speed(np.array([10, 20]), "KT"),
    ).value("DEG")
    assert r[0] == 225
    r = meteorology.drct(
        datatypes.speed(-10, "KT"), datatypes.speed(10, "KT")
    ).value("DEG")
    assert r == 135
    r = meteorology.drct(
        datatypes.speed(-10, "KT"), datatypes.speed(-10, "KT")
    ).value("DEG")
    assert r == 45
    r = meteorology.drct(
        datatypes.speed(10, "KT"), datatypes.speed(-10, "KT")
    ).value("DEG")
    assert r == 315


def test_windchill():
    """Wind Chill Conversion"""
    temp = datatypes.temperature(0, "F")
    sknt = datatypes.speed(30, "MPH")
    val = meteorology.windchill(temp, sknt).value("F")
    assert abs(val - -24.50) < 0.01


def test_dewpoint_from_pq():
    """ See if we can produce dew point from pressure and mixing ratio """
    p = datatypes.pressure(1013.25, "MB")
    mr = datatypes.mixingratio(0.012, "kg/kg")
    dwpk = meteorology.dewpoint_from_pq(p, mr)
    assert abs(dwpk.value("C") - 16.84) < 0.01


def test_dewpoint():
    """ test out computation of dew point """
    for t0, r0, a0 in [[80, 80, 73.42], [80, 20, 35.87]]:
        t = datatypes.temperature(t0, "F")
        rh = datatypes.humidity(r0, "%")
        dwpk = meteorology.dewpoint(t, rh)
        assert abs(dwpk.value("F") - a0) < 0.01


def test_heatindex():
    """ Test our heat index calculations """
    t = datatypes.temperature(80.0, "F")
    td = datatypes.temperature(70.0, "F")
    hdx = meteorology.heatindex(t, td)
    assert abs(hdx.value("F") - 83.93) < 0.01

    t = datatypes.temperature(30.0, "F")
    hdx = meteorology.heatindex(t, td)
    assert abs(hdx.value("F") - 30.00) < 0.01


def test_uv():
    """ Test calculation of uv wind components """
    speed = datatypes.speed([10], "KT")
    mydir = datatypes.direction([0], "DEG")
    u, v = meteorology.uv(speed, mydir)
    assert u.value("KT") == 0.0
    assert v.value("KT") == -10.0

    speed = datatypes.speed([10, 20, 15], "KT")
    mydir = datatypes.direction([90, 180, 135], "DEG")
    u, v = meteorology.uv(speed, mydir)
    assert u.value("KT")[0] == -10
    assert v.value("KT")[1] == 20.0
    assert abs(v.value("KT")[2] - 10.6) < 0.1


def test_relh():
    """ Simple check of bad units in temperature """
    tmp = datatypes.temperature(24, "C")
    dwp = datatypes.temperature(24, "C")
    relh = meteorology.relh(tmp, dwp)
    assert relh.value("%") == 100.0

    tmp = datatypes.temperature(32, "C")
    dwp = datatypes.temperature(10, "C")
    relh = meteorology.relh(tmp, dwp)
    assert abs(25.79 - relh.value("%")) < 0.01

    tmp = datatypes.temperature(32, "C")
    dwp = datatypes.temperature(15, "C")
    relh = meteorology.relh(tmp, dwp)
    assert abs(35.81 - relh.value("%")) < 0.01

    tmp = datatypes.temperature(5, "C")
    dwp = datatypes.temperature(4, "C")
    relh = meteorology.relh(tmp, dwp)
    assert abs(93.24 - relh.value("%")) < 0.01
